class Utils:
    from boto3.dynamodb.types import TypeSerializer, TypeDeserializer
    # The serializer and deserializer can be static, even in a lambda,
    # since their classes will never change according to the user.
    _serializer = None
    _deserializer = None

    @property
    def serializer(self) -> TypeSerializer:
        if Utils._serializer is None:
            Utils._serializer = Utils.TypeSerializer()
        return Utils._serializer
    
    @property
    def deserializer(self) -> TypeDeserializer:
        if Utils._deserializer is None:
            Utils._deserializer = Utils.TypeDeserializer()
        return Utils._deserializer

    def python_to_dynamodb(self, python_object):
        dynamodb_dict = self.serializer.serialize(python_object)
        dynamodb_dict_keys = list(dynamodb_dict.keys())
        if len(dynamodb_dict_keys) == 1:
            # When a dict is passed trough the serializer, there will be one excess key. For example, if we pass it {"main": {"child": True}} we
            # will get {'M': {'main': {'M': {'child': {'BOOL': True}}}}} But DynamoDB is not expecting the first 'M' key (stands for Map).
            # In order to have safer code, instead of returning the dict['M'] we return the value at the first key of the dict.
            return dynamodb_dict[dynamodb_dict_keys[0]]
        elif len(dynamodb_dict_keys) == 0:
            return dynamodb_dict
        else:
            raise Exception(f"The number of keys of the dynamodb was superior to 1. This is NOT SUPPOSED TO HAPPEN !"
                            f"\nPlease submit an issue on the GitHub page of the framework right away (https://github.com/Robinson04/inoft_vocal_framework)."
                            f"\nThere is something really wrong on our side or on AWS side.")

    def dynamodb_to_python(self, dynamodb_object):
        return self.deserializer.deserialize(dynamodb_object)



import logging

from inoft_vocal_framework.safe_dict import SafeDict
from inoft_vocal_framework.skill_builder.inoft_skill_builder import InoftHandlersGroup


class AlexaAudioPlayerHandlers(InoftHandlersGroup):
    from inoft_vocal_framework.skill_builder.inoft_skill_builder import InoftRequestHandler

    def __init__(self, parent_handler, additional_pause_intents_names: list = None,
                 additional_resume_intent_names: list = None):

        self.onPauseHandler = self.OnPauseHandler(parent_handler=parent_handler, intents_names=additional_pause_intents_names)
        self.onResumeHandler = self.OnResumeHandler(parent_handler=parent_handler, intent_names=additional_resume_intent_names)

    class OnPauseHandler(InoftRequestHandler):
        def __init__(self, parent_handler, intents_names: list):
            super().__init__(parent_handler=parent_handler)
            default_intents_names = ["AMAZON.PauseIntent"]
            self.intents_names = default_intents_names + intents_names if isinstance(intents_names, list) else default_intents_names
            self.request_types = ["PlaybackController.PauseCommandIssued"]

        def can_handle(self) -> bool:
            if self.is_in_intent_names(self.intents_names) or self.alexa.is_in_request_types(self.request_types):
                return True

        def handle(self) -> dict:
            self.say("Pause Request Received ! Please customize this message.")

            last_played_item = self.alexa.audioplayer.get_last_played_item()
            if last_played_item is not None:
                offset_in_milliseconds = self.alexa.audioplayer.offset_in_milliseconds
                if isinstance(offset_in_milliseconds, (int, float)) and offset_in_milliseconds > 0:
                    last_played_item.offset_in_milliseconds = offset_in_milliseconds
                    self.alexa.audioplayer.memorize_audio_item(last_played_item)
                    # We change the offset in milliseconds of the last played item, then we
                    # memorize the item back, in order to save its new modified offset value.

            self.alexa.audioplayer.stop()
            return self.to_platform_dict()

    class OnResumeHandler(InoftRequestHandler):
        def __init__(self, parent_handler, intent_names: list):
            super().__init__(parent_handler=parent_handler)
            default_intent_names = ["AMAZON.ResumeIntent"]
            self.intent_names = default_intent_names + intent_names if isinstance(intent_names, list) else default_intent_names
            self.request_types = ["PlaybackController.PlayCommandIssued"]

        def can_handle(self) -> bool:
            if self.is_in_intent_names(self.intent_names) or self.alexa.is_in_request_types(self.request_types):
                return True

        def handle(self) -> dict:
            if (self.alexa.audioplayer.token is not None
            and self.alexa.audioplayer.player_activity == self.alexa.audioplayer.PLAYER_ACTIVITY_STOPPED):
                last_played_item = self.alexa.audioplayer.get_last_played_item()
                if last_played_item is not None:
                    # Normally when resuming an audio media with the audioplayer, we should have an id for the last played
                    # item, and found it in the items dict. Yet it is possible for the id or the item to not be present if
                    # there has been some kind of manual or automatic reset of the user data in the database, and not in the
                    # alexa memory that remember that he had interacted with an audio player.
                    self.alexa.audioplayer.play_item(audio_item=last_played_item)
                    logging.debug(f"Resuming audio file with id {last_played_item.identifier} and title {last_played_item.title}")
                    return self.to_platform_dict()
                else:
                    # You should respond something like this :
                    self.say("Unfortunately i have not been able to find the track/podcast you last played. " +
                             "It's strange, it never happens to me. In the mean time that i found what happened, " +
                             "i can play a track/podcast that i think you would like. Are you up to it ?")
                    # And then have your custom logic to make it happen.

                    raise Exception("NOT IMPLEMENTED - What happens if the user is trying to ask the AudioPlayer to resume, that the"
                                    "AudioPlayer object is found in the request, yet that the last item played by the user has not"
                                    "been found in its playing history. This can happen if you have cleared the users data.")
            else:
                last_played_item = self.alexa.audioplayer.get_last_played_item()
                # todo: finish

                raise Exception("NOT IMPLEMENTED - What happens if the user is trying to ask the AudioPlayer to resume, that the"
                                "AudioPlayer object is found in the request, yet that the last item played by the user has not"
                                "been found in its playing history. This can happen if you have cleared the users data.")


    class OnPossibleResumeHandler(InoftRequestHandler):
        def __init__(self, parent_handler, intent_names: list):
            super().__init__(parent_handler=parent_handler)
            self.intent_names = intent_names if isinstance(intent_names, list) else list()

        def can_handle(self) -> bool:
            if (self.alexa.audioplayer.token is not None
            and self.alexa.audioplayer.player_activity == self.alexa.audioplayer.PLAYER_ACTIVITY_STOPPED):
                return True

        def handle(self):
            if (self.alexa.audioplayer.token is not None
            and self.alexa.audioplayer.player_activity == self.alexa.audioplayer.PLAYER_ACTIVITY_STOPPED):
                pass

    class OnNextHandler(InoftRequestHandler):
        def __init__(self, parent_handler, intent_names: list):
            super().__init__(parent_handler=parent_handler)
            default_intent_names = ["AMAZON.NextIntent"]
            self.intent_names = default_intent_names + intent_names if isinstance(intent_names, list) else default_intent_names
            self.request_types = ["PlaybackController.NextCommandIssued"]

        def can_handle(self) -> bool:
            if self.is_in_intent_names(self.intent_names) or self.alexa.is_in_request_types(self.request_types):
                return True

        def handle(self):
            raise Exception(f"Not implemented. You must customize this handler.")

    class OnPreviousHandler(InoftRequestHandler):
        def __init__(self, parent_handler, intent_names: list):
            super().__init__(parent_handler=parent_handler)
            default_intent_names = ["AMAZON.PreviousIntent"]
            self.intent_names = default_intent_names + intent_names if isinstance(intent_names, list) else default_intent_names
            self.request_types = ["PlaybackController.PreviousCommandIssued"]

        def can_handle(self) -> bool:
            if self.is_in_intent_names(self.intent_names) or self.alexa.is_in_request_types(self.request_types):
                return True

        def handle(self):
            raise Exception(f"Not implemented. You must customize this handler.")






from collections import Callable

from inoft_vocal_framework.dummy_object import DummyObject
from inoft_vocal_framework.platforms_handlers.dialogflow_v1.request import Request
from inoft_vocal_framework.platforms_handlers.dialogflow_v1.response import Response, Image, ImageDisplayOptions
from inoft_vocal_framework.safe_dict import SafeDict
from json import loads as json_loads


class SurfaceCapabilities:
    CAN_PLAY_AUDIO_MEDIA = "actions.capability.MEDIA_RESPONSE_AUDIO"
    HAS_AN_OUTPUT_SPEAKER = "actions.capability.AUDIO_OUTPUT"
    CAN_POTENTIALLY_USE_ACCOUNT_LINKING = "actions.capability.ACCOUNT_LINKING"
    CAN_ACCESS_WEB_BROWSER = "actions.capability.WEB_BROWSER"
    HAS_A_SCREEN = "actions.capability.SCREEN_OUTPUT"

    ALL_CAPABILITIES = [
        CAN_PLAY_AUDIO_MEDIA,
        HAS_AN_OUTPUT_SPEAKER,
        CAN_POTENTIALLY_USE_ACCOUNT_LINKING,
        CAN_ACCESS_WEB_BROWSER,
        HAS_A_SCREEN,
    ]


class DialogFlowHandlerInput(SurfaceCapabilities):
    from inoft_vocal_framework.platforms_handlers.handler_input import HandlerInput

    def __init__(self, parent_handler_input: HandlerInput):
        self.parent_handler_input = parent_handler_input
        self.request = Request()
        self.response = Response()

        self._user_id = None
        self._session_id = None
        self._is_new_session = None
        self._simple_session_user_data = None

    def _get_user_persistent_stored_data(self):
        unprocessed_user_stored_data = self.request.originalDetectIntentRequest.payload.user.userStorage
        if isinstance(unprocessed_user_stored_data, str) and unprocessed_user_stored_data.replace(" ", "") != "":
            from unicodedata import normalize as unicode_normalize
            from ast import literal_eval
            try:
                user_stored_data = literal_eval(unicode_normalize("NFD", unprocessed_user_stored_data))
                if isinstance(user_stored_data, dict):
                    return user_stored_data
                else:
                    print(f"Non-crashing error, the user_stored_data has been retrieved and converted from a string, but was but {type(user_stored_data)} and need to be a dict")
                    return None
            except Exception as e:
                print(f"Error while processing the user_persistent_data. Non-crashing but returning None : {e}")
        return None

    @property
    def user_id(self):
        persistent_data = self._get_user_persistent_stored_data()
        if isinstance(persistent_data, dict) and "userId" in persistent_data.keys():
            return persistent_data["userId"]
        else:
            return None

    @property
    def session_id(self) -> str:
        if self._session_id is None:
            self._session_id = self.request.session
        return self._session_id

    @property
    def is_new_session(self) -> bool:
        if self._is_new_session is None:
            self._is_new_session = True if self.request.originalDetectIntentRequest.payload.conversation.type == "NEW" else False
        return self._is_new_session

    @property
    def simple_session_user_data(self) -> SafeDict:
        for output_context in self.request.queryResult.outputContexts:
            if isinstance(output_context, dict) and "name" in output_context.keys() and "parameters" in output_context.keys():
                all_texts_after_slash = output_context["name"].split("/")
                last_text_after_slash = all_texts_after_slash[len(all_texts_after_slash) - 1]
                if str(last_text_after_slash).lower() == "sessiondata":
                    # We lower the text, to make sure that it will work even if the cases have been lowered. Because for some reasons,
                    # google is lowering the keys, so even if the key in the framework os sessionData, google might return sessiondata.
                    parameters_stringed_dict_or_dict = output_context["parameters"]
                    if parameters_stringed_dict_or_dict is not None:
                        if isinstance(parameters_stringed_dict_or_dict, str):
                            parameters_stringed_dict_or_dict = json_loads(parameters_stringed_dict_or_dict)
                        if isinstance(parameters_stringed_dict_or_dict, dict):
                            # The data key contains an stringed dictionary of the data we are interested by.
                            if "data" in parameters_stringed_dict_or_dict.keys():
                                parameters_stringed_dict_or_dict = parameters_stringed_dict_or_dict["data"]

                            if isinstance(parameters_stringed_dict_or_dict, str):
                                parameters_stringed_dict_or_dict = json_loads(parameters_stringed_dict_or_dict)
                            if isinstance(parameters_stringed_dict_or_dict, dict):
                                self._simple_session_user_data = SafeDict(parameters_stringed_dict_or_dict)
                            else:
                                self._simple_session_user_data = SafeDict()
                        else:
                            raise Exception(f"parameters_stringed_dict_or_dict was nto None, not a str, dict and could"
                                            f"not be json converted to a dict : {parameters_stringed_dict_or_dict}")

        if not isinstance(self._simple_session_user_data, SafeDict):
            self._simple_session_user_data = SafeDict()
        return self._simple_session_user_data

    def need_capabilities(self, capability_item_or_list):
        def check_if_capability_present_in_request(capability_to_check: str) -> bool:
            for capability_request_item in self.request.originalDetectIntentRequest.payload.surface.capabilities:
                if isinstance(capability_request_item, dict) and "name" in capability_request_item.keys():
                    if capability_request_item["name"] == capability_to_check:
                        return True
            return False

        if isinstance(capability_item_or_list, str):
            if (capability_item_or_list not in self.ALL_CAPABILITIES
            or check_if_capability_present_in_request(capability_to_check=capability_item_or_list) is False):
                return DummyObject()
            else:
                return self
        elif isinstance(capability_item_or_list, list):
            for capability_item in capability_item_or_list:
                if (capability_item not in self.ALL_CAPABILITIES
                or check_if_capability_present_in_request(capability_to_check=capability_item) is False):
                    return DummyObject()
            return self
        else:
            raise Exception(f"The variable passed to the need_capabilities function must be of type {list} or {str}"
                            f"but was {capability_item_or_list} of type {type(capability_item_or_list)}")

    def is_option_select_request(self) -> bool:
        return self.request.is_option_select_request()

    def selected_option_identifier(self) -> str:
        return self.request.selected_option_identifier()

    def is_launch_request(self) -> bool:
        return self.request.is_launch_request()

    def is_in_intent_names(self, intent_names_list) -> bool:
        if not isinstance(intent_names_list, list):
            if isinstance(intent_names_list, str):
                intent_names_list = [intent_names_list]
            else:
                raise Exception(f"The intent_names_list must be a list or a str in order to be converted to a list,"
                                f"but it was a {type(intent_names_list)} object : {intent_names_list}")

        return self.request.is_in_intent_names(intent_names_list=intent_names_list)

    def say(self, text_or_ssml: str) -> None:
        self.response.say(text_or_ssml=text_or_ssml)

    def reprompt(self, text_or_ssml: str) -> None:
        self.response.say_reprompt(text_or_ssml=text_or_ssml)

    def show_suggestion_chips(self, chips_titles: list):
        self.response.add_suggestions_chips(chips_titles=chips_titles)

    def show_basic_card(self, title: str = None, subtitle: str = None, content_formatted_text: str = None,
                        image: Image = None, image_display_options: ImageDisplayOptions = None, buttons: list = None) -> None:

        if content_formatted_text is None and image is None:
            raise Exception("A google assistant basic card cannot have both the content_formatted_text and the image variable as None")

        from inoft_vocal_framework.platforms_handlers.dialogflow_v1.response import BasicCard
        basic_card_object = BasicCard(title=title, subtitle=subtitle, formatted_text=content_formatted_text,
                                      image=image, image_display_options=image_display_options, buttons=buttons)
        self.response.add_response_item_to_show(item_object=basic_card_object)

    def show_browse_carousel_item(self, title: str, url_to_open_on_click: str, description: str = None,
                                  image_url: str = None, image_accessibility_text: str = None, footer: str = None):

        self.response.add_item_to_browse_carousel(title=title, url_to_open_on_click=url_to_open_on_click, description=description,
                                                  image_url=image_url, image_accessibility_text=image_accessibility_text, footer=footer)

    def _save_interaction_options_callback_function(self, on_select_callback_function: Callable, identifier_key: str):
        self.parent_handler_input.save_callback_function_to_database(callback_functions_key_name="interactivity_callback_functions",
            callback_function=on_select_callback_function, identifier_key=identifier_key)

    def show_interactive_list_item(self, identifier_key: str, on_select_callback: Callable, title: str, description: str = None,
                                   image_url: str = None, image_accessibility_text: str = None):

        self.response.add_interactive_list_item_to_system_intent(identifier_key=identifier_key, item_title=title, item_description=description,
                                                                 item_image_url=image_url, item_image_accessibility_text=image_accessibility_text)
        self._save_interaction_options_callback_function(on_select_callback_function=on_select_callback, identifier_key=identifier_key)

    def show_interactive_carousel_item(self, identifier_key: str, on_select_callback: Callable, title: str, description: str,
                                       image_url: str = None, image_accessibility_text: str = None):

        is_first_interactive_element = self.response.add_interactive_carousel_item_to_system_intent(identifier_key=identifier_key,
            item_title=title, item_description=description, item_image_url=image_url, item_image_accessibility_text=image_accessibility_text)
        self._save_interaction_options_callback_function(on_select_callback_function=on_select_callback, identifier_key=identifier_key)


    def play_audio(self, mp3_file_url: str, name: str, description: str = None,
                   icon_image_url: str = None, icon_accessibility_text: str = None,
                   override_default_end_session: bool = False):

        self.response.add_item_to_audio_media_response(mp3_file_url=mp3_file_url, name=name, description=description,
           icon_image_url=icon_image_url, icon_accessibility_text=icon_accessibility_text)

        if override_default_end_session is False:
            self.parent_handler_input.end_session()



from inoft_vocal_framework import InoftSkill, InoftRequestHandler, InoftStateHandler, InoftDefaultFallback
from messages import *
import os


class LaunchRequestHandler(InoftRequestHandler):
    def can_handle(self) -> bool:
        return self.is_launch_request()

    def handle(self):
        self.say(MSGS_WELCOME.pick())
        self.persistent_memorize("hasLaunchedOnce", True)
        return self.to_platform_dict()

class LaunchStateHandler(InoftStateHandler):
    def handle(self):
        if YesHandler(self).can_handle():
            self.say("The user said yes !")
            return self.to_platform_dict()

        elif NoHandler(self).can_handle():
            self.say("Why did you said no ?")
            return self.to_platform_dict()

    def fallback(self):
        self.say("You can say Yes or No and that's it")
        return self.to_platform_dict()

class YesHandler(InoftRequestHandler):
    def can_handle(self) -> bool:
        return self.is_in_intent_names(["AMAZON.YesIntent", "OkConfirmation"])

    def handle(self):
        self.say(f"Why did you said Yes even if you are not in a state handler ?")
        return self.to_platform_dict()

class NoHandler(InoftRequestHandler):
    def can_handle(self) -> bool:
        return self.is_in_intent_names(["AMAZON.NoIntent"])

    def handle(self):
        self.say(f"Why did you said No even if you are not in a state handler ?")
        return self.to_platform_dict()

class DefaultFallback(InoftDefaultFallback):
    def handle(self):
        self.say("I have no idea what you want to say")
        return self.to_platform_dict()


def lambda_handler(event, context):
    skill_builder = InoftSkill(settings_yaml_filepath=os.path.join(os.path.dirname(os.path.abspath(__file__)), "app_settings.yaml"))
    skill_builder.add_request_handler(LaunchRequestHandler)
    skill_builder.add_request_handler(YesHandler)
    skill_builder.add_request_handler(NoHandler)
    skill_builder.add_state_handler(LaunchStateHandler)
    skill_builder.set_default_fallback_handler(DefaultFallback)
    return skill_builder.handle_any_platform(event=event, context=context)

if __name__ == "__main__":
    from inoft_vocal_framework.platforms_handlers.simulator.simulator_core import Simulator
    event_, context_ = Simulator(event_type="google-assistant-v1_start-request").get_event_and_context()
    print(f"\n\nFinal Output : {lambda_handler(event=event_, context=context_)}")




from json import dumps as json_dumps

from inoft_vocal_framework.exceptions import raise_if_variable_not_expected_type
from inoft_vocal_framework.platforms_handlers.nested_object_to_dict import NestedObjectToDict

# todo: add select carousel
# todo: add select list

class OpenUrlAction:
    json_key = "openUrlAction"

    def __init__(self, url: str):
        self.url = url

    @property
    def url(self) -> str:
        return self._url

    @url.setter
    def url(self, url: str) -> None:
        if not isinstance(url, str):
            raise Exception(f"url was type {type(url)} which is not valid value for his parameter.")
        self._url = url

class Image:
    json_key = "image"

    def __init__(self, image_url: str, accessibility_text: str = None):
        self.url = image_url
        if accessibility_text is not None:
            # The accessibility_text is not required
            self.accessibilityText = accessibility_text
        else:
            self._accessibilityText = None

    @property
    def url(self) -> str:
        return self._url

    @url.setter
    def url(self, url: str) -> None:
        if not isinstance(url, str):
            raise Exception(f"url was type {type(url)} which is not valid value for his parameter.")
        self._url = url

    @property
    def accessibilityText(self) -> str:
        return self._accessibilityText

    @accessibilityText.setter
    def accessibilityText(self, accessibilityText: str) -> None:
        if accessibilityText is not None and not isinstance(accessibilityText, str):
            raise Exception(f"accessibilityText was type {type(accessibilityText)} which is not valid value for his parameter.")
        self._accessibilityText = accessibilityText


class Icon(Image):
    json_key = "icon"
    # todo: make sure that when a class inherite from another, we use the icon json_key and not the image key

    def __init__(self, image_url: str, accessibility_text: str = None):
        super().__init__(image_url=image_url, accessibility_text=accessibility_text)

class ImageDisplayOptions:
    json_key = "imageDisplayOptions"

    cropped_type = "CROPPED"
    available_options_types = [cropped_type]

    def __init__(self, option_type: str):
        self.option_type = option_type

    @property
    def option_type(self) -> str:
        return self._option_type

    @option_type.setter
    def option_type(self, option_type: str) -> None:
        if not isinstance(option_type, str):
            raise Exception(f"option_type was type {type(option_type)} which is not valid value for his parameter.")
        if option_type not in self.available_options_types:
            raise Exception(f"The option_type {option_type} was not a valid display options type.")
        self._option_type = option_type

class Button:
    json_key = "button"

    def __init__(self):
        self._title = str()
        self._openUrlAction = str()

    @property
    def title(self) -> str:
        return self._title

    @title.setter
    def title(self, title: str) -> None:
        if not isinstance(title, str):
            raise Exception(f"title was type {type(title)} which is not valid value for his parameter.")
        self._title = title

    @property
    def openUrlAction(self) -> str:
        return self._openUrlAction

    @openUrlAction.setter
    def openUrlAction(self, openUrlAction: str) -> None:
        if not isinstance(openUrlAction, str):
            raise Exception(f"openUrlAction was type {type(openUrlAction)} which is not valid value for his parameter.")
        self._openUrlAction = openUrlAction

class BasicCard:
    json_key = "basicCard"

    def __init__(self, title: str, subtitle: str, formatted_text: str = None, buttons: list = None,
                 image: Image = None, image_display_options: ImageDisplayOptions = None):

        self._title = title
        self._subtitle = subtitle
        self._formattedText = formatted_text
        self._buttons = buttons
        self._image = image
        self._imageDisplayOptions = image_display_options

    @property
    def title(self) -> str:
        return self._title

    @title.setter
    def title(self, title: str) -> None:
        if not isinstance(title, str):
            raise Exception(f"title was type {type(title)} which is not valid value for his parameter.")
        self._title = title

    @property
    def subtitle(self) -> str:
        return self._subtitle

    @subtitle.setter
    def subtitle(self, subtitle: str) -> None:
        if not isinstance(subtitle, str):
            raise Exception(f"subtitle was type {type(subtitle)} which is not valid value for his parameter.")
        self._subtitle = subtitle

    @property
    def formattedText(self) -> str:
        return self._formattedText

    @formattedText.setter
    def formattedText(self, formattedText: str) -> None:
        if not isinstance(formattedText, str):
            raise Exception(f"formattedText was type {type(formattedText)} which is not valid value for his parameter.")
        self._formattedText = formattedText

    @property
    def buttons(self) -> list:
        return self._buttons

    @buttons.setter
    def buttons(self, buttons: list) -> None:
        if not isinstance(buttons, list):
            raise Exception(f"buttons was type {type(buttons)} which is not valid value for his parameter.")
        self._buttons = buttons

    @property
    def image(self) -> Image:
        return self._image

    @image.setter
    def image(self, image: Image) -> None:
        if not isinstance(image, Image):
            raise Exception(f"image was type {type(image)} which is not valid value for his parameter.")
        self._image = image

    @property
    def imageDisplayOptions(self):
        return self._imageDisplayOptions

    @imageDisplayOptions.setter
    def imageDisplayOptions(self, imageDisplayOptionType: str) -> None:
        if not isinstance(imageDisplayOptionType, str):
            raise Exception(f"imageDisplayOptionType was type {type(imageDisplayOptionType)} which is not valid value for his parameter.")
        self._imageDisplayOptions.option_type = imageDisplayOptionType

    def return_transformations(self) -> None:
        self._imageDisplayOptions = self._imageDisplayOptions if self._imageDisplayOptions is None else self._imageDisplayOptions.option_type()


# todo: check if an image/video url is valid and throw an exception if not
# todo: raise for google if more than 2 response items have been set
# todo: do not allow to use elements that the plateform cannot support (like a carousel need to be on a platform that support a web browser, like a phone)

class BrowseCarousel:
    json_key = "carouselBrowse"

    def __init__(self):
        self._items = list()
        self._first_item_used_keys = None

    @property
    def items(self) -> list:
        return self._items

    def add_item(self, title: str, url_to_open_on_click: str, description: str = None,
                 image_url: str = None, image_accessibility_text: str = None, footer: str = None):

        if len(self._items) < 10:
            # A carousel can have a maximum of 10 items
            item = self.Item(title=title, description=description, openUrlAction=OpenUrlAction(url=url_to_open_on_click), footer=footer,
                             image=None if image_url is None else Image(image_url=image_url, accessibility_text=image_accessibility_text))

            if self._first_item_used_keys is None:
                self._first_item_used_keys = ["title", "url_to_open_on_click"]
                if description is not None:
                    self._first_item_used_keys.append("description")
                if image_url is not None:
                    self._first_item_used_keys.append("image_url")
                if image_accessibility_text is not None:
                    self._first_item_used_keys.append("image_accessibility_text")
                if footer is not None:
                    self._first_item_used_keys.append("footer")
            else:
                if (("description" in self._first_item_used_keys if description is None else "description" not in self._first_item_used_keys)
                or ("image_url" in self._first_item_used_keys if image_url is None else "image_url" not in self._first_item_used_keys)
                or ("image_accessibility_text" in self._first_item_used_keys if image_accessibility_text is None else "image_accessibility_text" not in self._first_item_used_keys)
                or ("footer" in self._first_item_used_keys if footer is None else "footer" not in self._first_item_used_keys)):
                    raise Exception(f"All carousel items must have the same variable, to assure a tile consistency."
                                    f"The first item of the carousel had the following variables : {self._first_item_used_keys}"
                                    f"which are not the same as the last item you inserted : {item.__dict__}")

            self._items.append(item)
        else:
            print("The browse carousel items limit of 10 has ben reached. Your new item has not been included.")

    def do_not_include(self):
        # If no items are present in the carousel, we do not include it
        if len(self._items) == 0:
            return True
        else:
            return False

    def return_transformations(self) -> None:
        if self.do_not_include() is False and len(self._items) == 1:
            # If the carousel has only one item, we duplicate it, because a carousel need at least 2 items
            first_item = self._items[0]
            if isinstance(first_item, self.Item):
                self._items.append(self.Item(title=first_item.title, description=first_item.description, image=first_item.image,
                                             openUrlAction=first_item.openUrlAction, footer=first_item.footer))
                print(f"Warning, a carousel had only one item, since we require two items, the only one has been duplicated.")
            else:
                raise Exception(f"The first_item of the carousel was not an instance of carousel item but {first_item}")

        del self._first_item_used_keys
        # The first_item_used_keys variable is only for backend practices and should not be
        # included in the response, so we delete the attributes on the return transformations.

    class Item:
        json_key = None

        def __init__(self, title: str, description: str = None, image: Image = None, openUrlAction: OpenUrlAction = None, footer: str = None):
            self.title = title
            self.description = description
            self.image = image
            self.openUrlAction = openUrlAction
            self.footer = footer

        @property
        def title(self) -> str:
            return self._title

        @title.setter
        def title(self, title: str) -> None:
            if not isinstance(title, str):
                raise Exception(f"title was type {type(title)} which is not valid value for his parameter.")
            self._title = title

        @property
        def description(self) -> str:
            return self._description

        @description.setter
        def description(self, description: str) -> None:
            if description is not None and not isinstance(description, str):
                raise Exception(f"description was type {type(description)} which is not valid value for his parameter.")
            self._description = description

        @property
        def image(self):
            return self._image

        @image.setter
        def image(self, image: Image) -> None:
            if image is not None and not isinstance(image, Image):
                raise Exception(f"image was type {type(image)} which is not valid value for his parameter.")
            self._image = image

        @property
        def openUrlAction(self):
            return self._openUrlAction

        @openUrlAction.setter
        def openUrlAction(self, openUrlAction: OpenUrlAction) -> None:
            if openUrlAction is not None and not isinstance(openUrlAction, OpenUrlAction):
                raise Exception(f"openUrlAction was type {type(openUrlAction)} which is not valid value for his parameter.")
            self._openUrlAction = openUrlAction

        @property
        def footer(self):
            return self._footer

        @footer.setter
        def footer(self, footer: str) -> None:
            if footer is not None and not isinstance(footer, str):
                raise Exception(f"footer was type {type(footer)} which is not valid value for his parameter.")
            self._footer = footer


class MediaResponse:
    json_key = "mediaResponse"
    key_media_type_audio = "AUDIO"
    available_media_types_keys = [key_media_type_audio]

    # todo: check that the user device has the capability to play audio files in a media response
    # todo: add follow up when audio file has finished playing

    def __init__(self):
        self._mediaType = None
        self._mediaObjects = list()

    @property
    def mediaType(self) -> str:
        return self._mediaType

    @mediaType.setter
    def mediaType(self, mediaType: str) -> None:
        if mediaType not in self.available_media_types_keys:
            raise Exception(f"The media_type {mediaType} was not found in the available media types list : {self.available_media_types_keys}")
        self._mediaType = mediaType

    @property
    def mediaObjects(self) -> list:
        return self._mediaObjects

    def add_audio_content(self, mp3_file_url: str, name: str, description: str = None,
                          icon_image_url: str = None, icon_accessibility_text: str = None):
        # todo: check validity of mp3 file, otherwise print a warning and do not add the item
        #  (and if it was the only item, do not include the media response)

        if self.mediaType is None:
            self.mediaType = self.key_media_type_audio
        else:
            if self.mediaType != self.key_media_type_audio:
                raise Exception(f"You cannot add an audio content object if you have already instantiated a media"
                                f"object by adding a different content that is of media type {self.mediaType}")

        self.mediaObjects.append(self.MediaObject(content_url=mp3_file_url, name=name, description=description,
                                                  icon=None if icon_image_url is None else Icon(image_url=icon_image_url,
                                                                                                accessibility_text=icon_accessibility_text)))

    class MediaObject:
        json_key = None

        def __init__(self, content_url: str, name: str, description: str = None, icon: Icon = None):
            self.contentUrl = content_url
            self.name = name
            self.description = description
            self.icon = icon

        @property
        def contentUrl(self) -> str:
            return self._contentUrl

        @contentUrl.setter
        def contentUrl(self, contentUrl: str) -> None:
            if contentUrl is not None and not isinstance(contentUrl, str):
                raise Exception(f"contentUrl was type {type(contentUrl)} which is not valid value for his parameter.")
            self._contentUrl = contentUrl

        @property
        def name(self) -> str:
            return self._name

        @name.setter
        def name(self, name: str) -> None:
            if name is not None and not isinstance(name, str):
                raise Exception(f"name was type {type(name)} which is not valid value for his parameter.")
            self._name = name

        @property
        def description(self):
            return self._description

        @description.setter
        def description(self, description: str) -> None:
            if description is not None and not isinstance(description, str):
                raise Exception(f"description was type {type(description)} which is not valid value for his parameter.")
            self._description = description

        @property
        def icon(self):
            return self._icon

        @icon.setter
        def icon(self, icon: Icon) -> None:
            if icon is not None and not isinstance(icon, Icon):
                raise Exception(f"icon was type {type(icon)} which is not valid value for his parameter.")
            self._icon = icon


class SystemIntent:
    json_key = "systemIntent"
    intent_type_option = "actions.intent.OPTION"
    available_intent_keys_types = [intent_type_option]
    element_type_list_select = "ListSelect"
    element_type_carousel_select = "CarouselSelect"
    available_elements_types_keys = [element_type_list_select, element_type_carousel_select]

    def __init__(self, element_type: str):
        self.intent = self.intent_type_option
        self._data = self.Data(element_type=element_type)

    @property
    def intent(self) -> str:
        return self._intent

    @intent.setter
    def intent(self, intent_type: str) -> None:
        if intent_type not in self.available_intent_keys_types:
            raise Exception(f"The key_type {intent_type} was not found in the intent keys types list : {self.available_intent_keys_types}")
        self._intent = intent_type

    class Data:
        json_key = "data"

        def __init__(self, element_type: str):
            if element_type == SystemIntent.element_type_list_select:
                self._type = "type.googleapis.com/google.actions.v2.OptionValueSpec"
                self._listSelect = self.ListSelect()
            elif element_type == SystemIntent.element_type_carousel_select:
                self._type = "type.googleapis.com/google.actions.v2.OptionValueSpec"
                self._carouselSelect = self.CarouselSelect()
            else:
                raise Exception(f"The element_type {element_type} is not a supported element type.")

        @property
        def type(self) -> str:
            return self._type

        @type.setter
        def type(self, type_key: str) -> None:
            if not isinstance(type_key, str):
                raise Exception(f"type was type {type(type_key)} which is not valid value for his parameter.")
            self._type = type_key

        def return_transformations(self):
            # The key that google is expecting is the @type key, yet since in python i cannot directly set a variable
            # with the @type name, we use a simple '_type' variable name, and in the return transformations, we create
            # a new var called '@type' with the vars dict to which we assign the value of the 'type' property (it points
            # to the '_type' variable) and then we can delete the '_type' variable to not include it in the output dict.
            vars(self)["@type"] = self.type
            del self._type

        class InteractiveListOrCarouselItem:
            json_key = None

            def __init__(self, identifier_key: str, title: str, description: str = None, image: Image = None):
                self._title = title
                self._description = description
                self._image = image
                self._optionInfo = self.OptionInfo(identifier_key=identifier_key)

            @property
            def title(self) -> str:
                return self._title

            @title.setter
            def title(self, title: str) -> None:
                raise_if_variable_not_expected_type(value=title, expected_type=str, variable_name="title")
                self._title = title

            @property
            def description(self) -> str:
                return self._description

            @description.setter
            def description(self, description: str) -> None:
                raise_if_variable_not_expected_type(value=description, expected_type=str, variable_name="description")
                self._description = description

            @property
            def image(self) -> Image:
                return self._image

            @image.setter
            def image(self, image: Image) -> None:
                raise_if_variable_not_expected_type(value=image, expected_type=Image, variable_name="image")
                self._image = image

            class OptionInfo:
                json_key = "optionInfo"

                def __init__(self, identifier_key: str):
                    self.key = identifier_key
                    import random  # todo: check if synonyms are required and add possibility to set them
                    self._synonyms = [f"{random.randint(0, 1000000000000000)}"]

                @property
                def key(self) -> str:
                    return self._key

                @key.setter
                def key(self, key: str) -> None:
                    raise_if_variable_not_expected_type(value=key, expected_type=str, variable_name="key")
                    self._key = key

                @property
                def synonyms(self) -> list:
                    return self._synonyms

                @synonyms.setter
                def synonyms(self, synonyms_list: list) -> None:
                    raise_if_variable_not_expected_type(value=synonyms_list, expected_type=list, variable_name="synonyms_list")
                    self._synonyms = synonyms_list

            @property
            def optionInfo(self) -> OptionInfo:
                return self._optionInfo

        class ListSelect:
            json_key = "listSelect"

            def __init__(self):
                self._title = str()
                self._items = list()

            def add_item(self, identifier_key: str, title: str, description: str = None,
                         image_url: str = None, image_accessibility_text: str = None):

                if len(self._items) < 30:
                    self.items.append(SystemIntent.Data.InteractiveListOrCarouselItem(
                        identifier_key=identifier_key, title=title, description=description,
                        image=None if image_url is None else Image(image_url=image_url, accessibility_text=image_accessibility_text)))
                else:
                    print("The list items limit of 30 has ben reached. Your new item has not been included.")

            def do_not_include(self):
                # If no items are present in the list, we do not include it
                if len(self.items) == 0:
                    return True
                else:
                    return False

            def return_transformations(self) -> None:
                if self.do_not_include() is False and len(self._items) == 1:
                    # If the carousel has only one item, we duplicate it, because a carousel need at least 2 items
                    first_item = self.items[0]
                    if isinstance(first_item, SystemIntent.Data.InteractiveListOrCarouselItem):
                        new_identifier_key = f"{first_item.optionInfo.key}.2"
                        self.items.append(SystemIntent.Data.InteractiveListOrCarouselItem(
                            identifier_key=new_identifier_key, title=f"2 - {first_item.title}",
                            description=first_item.description, image=first_item.image))

                        first_item.title = f"1 - {first_item.title}"
                        print(f"Warning, a list had only one item, since we require two items, the only one has been duplicated"
                              f"with a new id of {new_identifier_key} instead of dumbly duplicating the id {first_item.optionInfo.key}"
                              f"A '1 -' and a '2 -' have also been added to the elements titles, to make them both uniques.")
                    else:
                        raise Exception(f"The first_item of the list was not an instance of list item but {first_item}")

            @property
            def title(self) -> str:
                return self._title

            @title.setter
            def title(self, title: str) -> None:
                raise_if_variable_not_expected_type(value=title, expected_type=str, variable_name="title")
                self._title = title

            @property
            def items(self) -> list:
                return self._items

            @items.setter
            def items(self, items_list: list) -> None:
                raise_if_variable_not_expected_type(value=items_list, expected_type=list, variable_name="items_list")
                self._items = items_list

        @property
        def listSelect(self) -> ListSelect:
            if "_listSelect" not in vars(self):
                raise Exception(f"The SystemIntent was not set to a ListSelect. Did you tried to show different types"
                                f"of elements that run with a SystemIntent, in the same execution of your app ?")
            else:
                return self._listSelect

        class CarouselSelect:
            json_key = "carouselSelect"

            def __init__(self):
                self._items = list()

            def add_item(self, identifier_key: str, title: str, description: str,
                         image_url: str = None, image_accessibility_text: str = None):

                if len(self._items) < 10:
                    self.items.append(SystemIntent.Data.InteractiveListOrCarouselItem(
                        identifier_key=identifier_key, title=title, description=description,
                        image=None if image_url is None else Image(image_url=image_url, accessibility_text=image_accessibility_text)))
                else:
                    print("The carousel items limit of 10 has ben reached. Your new item has not been included.")

            def do_not_include(self):
                # If no items are present in the carousel, we do not include it
                if len(self.items) == 0:
                    return True
                else:
                    return False

            def return_transformations(self) -> None:
                if self.do_not_include() is False and len(self._items) == 1:
                    # If the carousel has only one item, we duplicate it, because a carousel need at least 2 items
                    first_item = self.items[0]
                    if isinstance(first_item, SystemIntent.Data.InteractiveListOrCarouselItem):
                        new_identifier_key = f"{first_item.optionInfo.key}.2"
                        self.items.append(SystemIntent.Data.InteractiveListOrCarouselItem(
                            identifier_key=new_identifier_key, title=f"2 - {first_item.title}",
                            description=first_item.description, image=first_item.image))

                        first_item.title = f"1 - {first_item.title}"
                        print(f"Warning, a list had only one item, since we require two items, the only one has been duplicated"
                              f"with a new id of {new_identifier_key} instead of dumbly duplicating the id {first_item.optionInfo.key}"
                              f"A '1 -' and a '2 -' have also been added to the elements titles, to make them both uniques.")
                    else:
                        raise Exception(f"The first_item of the list was not an instance of list item but {first_item}")

            @property
            def items(self) -> list:
                return self._items

            @items.setter
            def items(self, items_list: list) -> None:
                raise_if_variable_not_expected_type(value=items_list, expected_type=list, variable_name="items_list")
                self._items = items_list

        @property
        def carouselSelect(self) -> CarouselSelect:
            if "_carouselSelect" not in vars(self):
                raise Exception(f"The SystemIntent was not set to a CarouselSelect. Did you tried to show different types"
                                f"of elements that run with a SystemIntent, in the same execution of your app ?")
            else:
                return self._carouselSelect

    @property
    def data(self) -> Data:
        return self._data

class SimpleResponse:
    json_key = "simpleResponse"

    def __init__(self):
        self._textToSpeech = str()
        self._displayText = str()

    @property
    def textToSpeech(self):
        return self._textToSpeech

    @textToSpeech.setter
    def textToSpeech(self, text: str) -> None:
        if isinstance(text, str):
            self._textToSpeech = text
        else:
            raise Exception(f"The text was not a string object : {text}")

    @property
    def displayText(self):
        return self._textToSpeech

    @displayText.setter
    def displayText(self, text: str) -> None:
        if isinstance(text, str):
            self._displayText = text
        else:
            raise Exception(f"The text was not a string object : {text}")


class RichResponseInPayload:
    json_key = "richResponse"

    def __init__(self):
        self._items = list()
        self.suggestions = list()

    @property
    def items(self):
        return self._items

    def add_response_item(self, response_item_object):
        response_item_dict = NestedObjectToDict.get_dict_from_nested_object(
            object_to_process=response_item_object, key_names_identifier_objects_to_go_into=["json_key"])
        self._items.append(response_item_dict)

    def add_suggestion_chip(self, title: str):
        # todo: add external link suggestion chip (cannot be used on platforms without the actions.capability.WEB_BROWSER capability
        if not isinstance(title, str):
            raise Exception(f"The title variable must be of type str but was {type(title)}")
        if len(title) > 25:
            raise Exception(f"A suggestion chip title can have a maximum of 25 chars but {title} was {len(title)} chars")

        self.suggestions.append({"title": title})

class GoogleInPayload:
    json_key = "google"

    def __init__(self):
        self._expectUserResponse = True
        self.richResponse = RichResponseInPayload()
        self._userStorage = str()

    @property
    def expectUserResponse(self):
        return self._expectUserResponse

    @expectUserResponse.setter
    def expectUserResponse(self, expectUserResponse) -> None:
        if expectUserResponse is False or expectUserResponse is True:
            self._expectUserResponse = expectUserResponse
        else:
            raise Exception(f"expectUserResponse can only receive a True or False value, it received the following : {expectUserResponse}")

    @property
    def userStorage(self) -> str:
        return self._userStorage

    @userStorage.setter
    def userStorage(self, userStorage: str) -> None:
        if not isinstance(userStorage, str):
            raise Exception(f"userStorage was type {type(userStorage)} which is not valid value for his parameter.")
        self._userStorage = userStorage

class Payload:
    json_key = "payload"

    def __init__(self):
        self.google = GoogleInPayload()

    def to_dict(self) -> dict:
        return NestedObjectToDict.get_dict_from_nested_object(object_to_process=self,
                                                              key_names_identifier_objects_to_go_into=["json_key"])

class OutputContextItem:
    json_key = None
    session_data_name = "sessionData"

    def __init__(self, session_id: str, name: str, lifespanCount=999):
        self.name = f"{session_id}/contexts/{name}"
        self.lifespanCount = lifespanCount
        self._parameters = dict()

    def return_transformations(self) -> None:
        if "data" in self.parameters.keys():
            self.parameters["data"] = json_dumps(self._parameters["data"])

    def add_set_parameter(self, parameter_key: str, parameter_value=None):
        """
        :param parameter_key: str key for the parameters dict object, should not be an empty str
        :param parameter_value: any object, cannot be None
        :return: self
        """
        if parameter_value is not None and isinstance(parameter_key, str) and parameter_key != "":
            self._parameters[parameter_key] = parameter_value
        return self

    def add_set_session_attribute(self, parameter_key: str, parameter_value=None):
        if "data" not in self.parameters.keys():
            self.parameters["data"] = dict()

        if parameter_value is not None and isinstance(parameter_key, str) and parameter_key != "":
            self._parameters["data"][parameter_key] = parameter_value
        return self

    @property
    def parameters(self) -> dict:
        return self._parameters

    @parameters.setter
    def parameters(self, parameters: dict) -> None:
        if not isinstance(parameters, dict):
            raise Exception(f"parameters was type {type(parameters)} which is not valid value for his parameter.")
        self._parameters = parameters

class Response:
    json_key = "response"

    def __init__(self):
        self.payload = Payload()
        self.outputContexts = list()

    def say(self, text_or_ssml: str) -> None:
        # todo: allow to have 2 differents response in the same one, not just one
        output_response = SimpleResponse()
        output_response.textToSpeech = text_or_ssml
        self.payload.google.richResponse.add_response_item(output_response)

    def say_reprompt(self, text_or_ssml: str) -> None:
        # todo: finish the reprompt function
        return None
        output_response = SimpleResponse()
        output_response.textToSpeech = text_or_ssml
        self.payload.google.richResponse.add_response_item(output_response)

    def add_output_context_item(self, output_context_item: OutputContextItem) -> None:
        if isinstance(output_context_item, OutputContextItem):
            self.outputContexts.append(output_context_item)
        else:
            raise Exception(f"The output context item needed to be of instance {OutputContextItem} but was : {output_context_item}")

    def add_response_item_to_show(self, item_object) -> None:
        if not len(self.payload.google.richResponse.items) > 0:
            raise Exception("A google response take in consideration the order with which the elements have been added. "
                            "You cannot set a visual element as first element of the response. "
                            "Try to set a speech element with the .say() function before you add a visual element.")
        self.payload.google.richResponse.add_response_item(item_object)

    def add_suggestions_chips(self, chips_titles: list) -> None:
        if isinstance(chips_titles, list):
            for chip_title in chips_titles:
                self.payload.google.richResponse.add_suggestion_chip(title=chip_title)
        elif isinstance(chips_titles, str):
            self.payload.google.richResponse.add_suggestion_chip(title=chips_titles)

    def add_item_to_browse_carousel(self, title: str,  url_to_open_on_click: str, description: str = None,
                                    image_url: str = None, image_accessibility_text: str = None, footer: str = None):

        carousel_instance = None
        # todo: improve this loop so that we do not need to loop every time we add an item
        for response_item in self.payload.google.richResponse.items:
            if isinstance(response_item, BrowseCarousel):
                carousel_instance = response_item
                break
        if carousel_instance is None:
            carousel_instance = BrowseCarousel()
            self.payload.google.richResponse.items.append(carousel_instance)
        carousel_instance.add_item(title=title, url_to_open_on_click=url_to_open_on_click, description=description,
                                   image_url=image_url, image_accessibility_text=image_accessibility_text, footer=footer)

    def add_interactive_list_item_to_system_intent(self, identifier_key: str, item_title: str, item_description: str = None,
                                                   item_image_url: str = None, item_image_accessibility_text: str = None) -> bool:

        if "systemIntent" not in vars(self.payload.google):
            self.payload.google.systemIntent = SystemIntent(element_type=SystemIntent.element_type_list_select)
        elif self.payload.google.systemIntent.data.listSelect is None:
            # If the listSelect variable is None, it means that the SystemIntent do not represent a ListSelect,
            # and an exception will be thrown (in the data class itself, the below exception will never be called).
            raise Exception(f"A systemIntent as already been created. Make sure that you do not tried to show"
                            f"another interactive (or static element) that use a systemIntent somewhere else.")

        self.payload.google.systemIntent.data.listSelect.add_item(identifier_key=identifier_key, title=item_title, description=item_description,
                                                                  image_url=item_image_url, image_accessibility_text=item_image_accessibility_text)

    def add_interactive_carousel_item_to_system_intent(self, identifier_key: str, item_title: str, item_description: str,
                                                       item_image_url: str = None, item_image_accessibility_text: str = None) -> bool:
        """
        :return bool: True if the SystemIntent has been created and so we just set the first interactive element in the invocation and False otherwise
        """

        is_first_interactive_element_of_invocation = False
        if "systemIntent" not in vars(self.payload.google):
            self.payload.google.systemIntent = SystemIntent(element_type=SystemIntent.element_type_carousel_select)
            is_first_interactive_element_of_invocation = True
        elif self.payload.google.systemIntent.data.carouselSelect is None:
            # If the carouselSelect variable is None, it means that the SystemIntent do not represent a carouselSelect,
            # and an exception will be thrown (in the data class itself, the below exception will never be called).
            raise Exception(f"A systemIntent as already been created. Make sure that you do not tried to show"
                            f"another interactive (or static element) that use a systemIntent somewhere else.")

        self.payload.google.systemIntent.data.carouselSelect.add_item(identifier_key=identifier_key, title=item_title, description=item_description,
                                                                      image_url=item_image_url, image_accessibility_text=item_image_accessibility_text)
        return is_first_interactive_element_of_invocation

    def add_item_to_audio_media_response(self, mp3_file_url: str, name: str, description: str = None,
                                         icon_image_url: str = None, icon_accessibility_text: str = None):

        media_response_instance = None
        # todo: improve this loop so that we do not need to loop every time we add an item
        for response_item in self.payload.google.richResponse.items:
            if isinstance(response_item, MediaResponse):
                media_response_instance = response_item
                break
        if media_response_instance is None:
            media_response_instance = MediaResponse()
            self.payload.google.richResponse.items.append(media_response_instance)
        media_response_instance.add_audio_content(mp3_file_url=mp3_file_url, name=name, description=description,
                                                  icon_image_url=icon_image_url, icon_accessibility_text=icon_accessibility_text)

    def to_dict(self) -> dict:
        return NestedObjectToDict.get_dict_from_nested_object(object_to_process=self,
                                                              key_names_identifier_objects_to_go_into=["json_key"])[self.json_key]


if __name__ == "__main__":
    NestedObjectToDict.get_dict_from_nested_object(Payload(), ["json_key"])


import os
import string
import boto3
import click
from pathlib import Path
from random import choice
import inoft_vocal_framework
from inoft_vocal_framework.cli.aws_utils import is_valid_bucket_name, BUCKET_NAMING_MSG


BOTO3_CONFIG_DOCS_URL = "https://boto3.readthedocs.io/en/latest/guide/quickstart.html#configuration"

def _get_from_name_or_ask_to_select_template(selected_template_name: str = None) -> dict:
    current_folder_path = Path(os.path.dirname(os.path.abspath(__file__)))
    templates_folder_path = os.path.join(current_folder_path.parent, "templates")
    invalid_folder_names = ["__pycache__"]

    available_templates = dict()
    for template_dir_name in os.listdir(templates_folder_path):
        if template_dir_name not in invalid_folder_names:
            if os.path.isdir(os.path.join(templates_folder_path, template_dir_name)):
                current_template_dir_path = os.path.join(templates_folder_path, template_dir_name)
                files_names_in_current_template_dir = os.listdir(current_template_dir_path)

                if not len(files_names_in_current_template_dir) > 0:
                    print(f"That's strange. The template folder at {current_template_dir_path} "
                          f"did not contain any file, and so cannot be used.")
                else:
                    paths_to_files_to_copy = list()
                    for filename in files_names_in_current_template_dir:
                        if filename not in ["infos.yaml", "infos.json"]:
                            filepath = os.path.join(current_template_dir_path, filename)
                            if os.path.isfile(filepath):
                                paths_to_files_to_copy.append(filepath)

                    if not len(paths_to_files_to_copy) > 0:
                        print(f"That's strange. The template folder at {current_template_dir_path} "
                              f"did not contain any file that could be copied, and so cannot be used.")
                    else:
                        current_template_infos_dict = None

                        if "infos.yaml" in files_names_in_current_template_dir:
                            from inoft_vocal_framework.utils.general import load_yaml
                            current_template_infos_dict = load_yaml(filepath=os.path.join(current_template_dir_path, "infos.yaml"))
                        elif "infos.json" in files_names_in_current_template_dir:
                            from inoft_vocal_framework.utils.general import load_json
                            current_template_infos_dict = load_json(filepath=os.path.join(current_template_dir_path, "infos.json"))

                        if isinstance(current_template_infos_dict, dict):
                            if "name" in current_template_infos_dict.keys() and isinstance(current_template_infos_dict["name"], str):
                                available_templates[current_template_infos_dict["name"]] = {"pathsToFilesToCopy": paths_to_files_to_copy}
                                continue

                        available_templates[template_dir_name] = {"pathsToFilesToCopy": paths_to_files_to_copy}
                        # If the no infos dict has been created or that no name variable has been found in the infos, we will not have
                        # triggered the continue keyword, and so we will reach this stage where we add the template with its folder name.

    if selected_template_name is None:
        selected_template_name = click.prompt("Type the name of the template you would like to use, the followings are available :",
                                              type=click.Choice(available_templates.keys()))
        # todo: improve that, i would like for the user to select it with a list select (like in PyInquirer)

    if selected_template_name in available_templates.keys():
        return available_templates[selected_template_name]
    else:
        return None

def new(template_name: str = None, project_folderpath: str = None):
    current_folder_path = Path(os.path.dirname(os.path.abspath(__file__)))
    selected_template_dict = _get_from_name_or_ask_to_select_template(selected_template_name=template_name)
    selected_templates_files_to_copy = selected_template_dict["pathsToFilesToCopy"]

    if project_folderpath is None:
        while True:
            project_folderpath = click.prompt(text="To which folder location would like to put the template ?",
                                              default=Path(os.path.dirname(os.path.realpath(inoft_vocal_framework.__file__))).parent)
            if not os.path.isdir(project_folderpath):
                print(f"No folder has been found at {click.style(text=project_folderpath, bold=True, fg='blue')}."
                      f"\nPlease write a valid folder path, or exit the CLI by using CTRL+C")
            else:
                break

            """if click.confirm("You now need to select a folder where your project (and your template) will be located. Type y"):
                from inoft_vocal_framework.cli.gui_handlers import select_folder
                project_folderpath = select_folder(title="Folder for your project", initial_dir=current_folder_path.parents[1])
                break
                # todo: find a way to make the gui works (right now, it just freeze the script and the folderpicker window never opens)
            else:
                if click.confirm("Do you want to select a different template ?"):
                    selected_template_dict = _get_from_name_or_ask_to_select_template()
                    selected_templates_files_to_copy = selected_template_dict["pathsToFilesToCopy"]
                else:
                    print("We do not know how to help you."
                          "\nIf you have questions on how the framework works, please visit or GitHub page : " +
                          click.style("https://github.com/Robinson04/inoft_vocal_framework", fg="cyan", bold=True))"""

    Path(project_folderpath).mkdir(exist_ok=True)

    for src_file_path in selected_templates_files_to_copy:
        from shutil import copy
        destination_filepath = os.path.join(project_folderpath, Path(src_file_path).name)
        if os.path.isfile(destination_filepath):
            if click.confirm(f"A file already exist at {destination_filepath}. Do you wish to keep the existing file and not replace it ?"):
                continue
        click.echo(f"Copying {src_file_path} to {destination_filepath}")
        copy(src_file_path, destination_filepath)
    click.echo("Template copying completed.")

    """
    # Detect AWS profiles and regions
    session = boto3.session.Session()
    profile_names = session.available_profiles

    click.echo("\nAWS Lambda and API Gateway are only available in certain regions. "
               "Let's check to make sure you have a profile set up in one that will work.")

    if not profile_names:
        click.echo(f"We couldn't find an AWS profile to use. Before using the Inoft Vocal Framework, you'll need to set one up."
                   f"\nSee here for more info: {click.style(BOTO3_CONFIG_DOCS_URL, fg='blue', underline=True)}")
        return None
    elif len(profile_names) == 1:
        used_profile_name = profile_names[0]
        click.echo(f"Okay, using profile '{click.style(used_profile_name, bold=True)}' !")
    else:
        used_profile_name = click.prompt(f"We found the following profiles type the one you would like to use", type=click.Choice(profile_names))

    # Create Bucket
    click.echo("\nYour skills deployments will need to be uploaded to a " + click.style("private S3 bucket", bold=True) + ".")
    click.echo("If you don't have a bucket yet, we'll create one for you too.")
    default_bucket_name = f"inoftskill-{''.join(choice(string.ascii_lowercase + string.digits) for _ in range(9))}"

    bucket_name = click.prompt("What is/will be the name of your bucket ? Default :", default=default_bucket_name)
    while True:
        bucket_name = click.prompt("What is/will be the name of your bucket ? Default :", default=default_bucket_name)

        if is_valid_bucket_name(bucket_name):
            break

        click.echo(click.style("Invalid bucket name!", bold=True))
        click.echo("S3 buckets must be named according to the following rules:")
        click.echo(BUCKET_NAMING_MSG)

    # todo: add multi-regions deployement

    # The given environment name
    zappa_settings = {
        env: {
            'profile_name': profile_name,
            's3_bucket': bucket,
            'runtime': get_venv_from_python_version(),
            'project_name': self.get_project_name()
        }
    }

    if profile_region:
        zappa_settings[env]['aws_region'] = profile_region


    import json as json # json is fine for loading, not fine for writing.
    zappa_settings_json = json.dumps(zappa_settings, sort_keys=True, indent=4)

    click.echo("\nOkay, here's your " + click.style("zappa_settings.json", bold=True) + ":\n")
    click.echo(click.style(zappa_settings_json, fg="yellow", bold=False))

    confirm = input("\nDoes this look " + click.style("okay", bold=True, fg="green")  + "? (default 'y') [y/n]: ") or 'yes'
    if confirm[0] not in ['y', 'Y', 'yes', 'YES']:
        click.echo("" + click.style("Sorry", bold=True, fg='red') + " to hear that! Please init again.")
        return

    # Write
    with open("zappa_settings.json", "w") as zappa_settings_file:
        zappa_settings_file.write(zappa_settings_json)
    """

    click.echo("\nPour en savoir plus sur le framework, rendez-vous sur notre page " + click.style("GitHub", bold=True) +
               " ici : " + click.style("https://github.com/Robinson04/inoft_vocal_framework", fg="cyan", bold=True))
    click.echo("\nExcellente journée ! ;)")
    click.echo(" ~ " + click.style("Robinson Labourdette d'Inoft", bold=True) + "!")

    return




import logging
from abc import abstractmethod
from json import dumps as json_dumps

from inoft_vocal_framework.dummy_object import DummyObject
from inoft_vocal_framework.exceptions import raise_if_value_not_in_list
from inoft_vocal_framework.platforms_handlers.endpoints_providers.providers import LambdaResponseWrapper
from inoft_vocal_framework.platforms_handlers.handler_input import HandlerInput, HandlerInputWrapper
from inoft_vocal_framework.platforms_handlers.nested_object_to_dict import NestedObjectToDict
from inoft_vocal_framework.safe_dict import SafeDict
from inoft_vocal_framework.skill_builder.skill_settings import Settings

class InoftHandler(HandlerInputWrapper):
    @abstractmethod
    def handle(self) -> dict:
        """Handles the Request inside handler input and provides a Response for dispatcher to return.
        :return: Response for the dispatcher to return or None
        :rtype: Union[Response, None]
        """
        raise NotImplementedError

    @abstractmethod
    def handle_resume(self) -> dict:
        # todo: make the handle resume function functionnal for the handler (cannot use a chain, but need to
        # use a class path, that will be saved in the database)
        print(f"Resuming an user session, but no logic has been found in the handle_resume function, defaulting to the handle function")

class InoftRequestHandler(HandlerInputWrapper):  # (HandlerInput):
    @abstractmethod
    def can_handle(self) -> bool:
        """ Returns true if Request Handler can handle the Request inside Handler Input.
        :return: Boolean value that tells the dispatcher if the current request can be handled by this handler.
        :rtype: bool
        """
        raise NotImplementedError

    @abstractmethod
    def handle(self) -> dict:
        """Handles the Request inside handler input and provides a Response for dispatcher to return.
        :return: Response for the dispatcher to return or None
        :rtype: Union[Response, None]
        """
        raise NotImplementedError

    @abstractmethod
    def handle_resume(self) -> dict:
        print(f"Resuming an user session, but no logic has been found in the handle_resume function, defaulting to the handle function")

class InoftStateHandler(HandlerInputWrapper):  # (HandlerInput):
    # todo: make it possible for a state or request handler to be a nested class inside another one
    @abstractmethod
    def handle(self) -> dict:
        """Handles the Request inside handler input and provides a Response for dispatcher to return.
        :return: Response for the dispatcher to return or None
        :rtype: Union[Response, None]
        """
        raise NotImplementedError

    @abstractmethod
    def fallback(self) -> dict:
        """ Handler if no response has been gotten from the handle method.
        :return: Response for the dispatcher to return or None
        :rtype: Union[Response, None]
        """
        raise NotImplementedError

    @abstractmethod
    def handle_resume(self):
        print(f"Resuming an user session, but no logic has been found in the handle_resume function, defaulting to the handle function")

class InoftDefaultFallback(HandlerInputWrapper):
    @abstractmethod
    def handle(self) -> dict:
        raise NotImplementedError

class InoftHandlersGroup:
    @abstractmethod
    def __getattr__(self, item):
        self_vars = vars(self)
        if item in self_vars:
            return self_vars[item]
        else:
            return DummyObject()

    def handle(self) -> dict:
        for var_key, var_object in vars(self).items():
            if InoftRequestHandler in list(var_object.__class__.__bases__):
                if var_object.can_handle() is True:
                    handler_output = var_object.handle()
                    if handler_output is not None:
                        return handler_output


class InoftSkill:
    def __init__(self, settings_yaml_filepath=None, settings_json_filepath=None):

        self.settings = Settings()
        if settings_yaml_filepath is not None and settings_json_filepath is not None:
            raise Exception(f"You cannot specify multiple settings files. Please specify only one")
        elif settings_yaml_filepath is None and settings_json_filepath is None:
            raise Exception(f"Please specify a yaml or json settings file with the settings_yaml_filepath arg or settings_json_filepath")
        elif settings_yaml_filepath is not None:
            self.settings.load_yaml(settings_file=settings_yaml_filepath)
        elif settings_json_filepath is not None:
            self.settings.load_json(settings_file=settings_json_filepath)

        self._request_handlers_chain = dict()
        self._state_handlers_chain = dict()
        self._default_fallback_handler = None
        self._handler_input = HandlerInput()

    def add_request_handler(self, request_handler_instance_or_class) -> None:
        if request_handler_instance_or_class is not None:
            try:
                if isinstance(request_handler_instance_or_class, type):
                    # If the variable is a class object we create an instance of the class
                    handler_bases_parent_classes = request_handler_instance_or_class.__bases__
                    request_handler_instance_or_class = request_handler_instance_or_class()
                else:
                    # If the variable is a class instance
                    handler_bases_parent_classes = request_handler_instance_or_class.__class__.__bases__

                if InoftRequestHandler in handler_bases_parent_classes:
                    request_handler_instance_or_class.handler_input = self.handler_input
                    # We set a reference to the skill handler_input in each handler so that it can use it with its HandlerInputWrapper
                    self.request_handlers_chain[request_handler_instance_or_class.__class__.__name__] = request_handler_instance_or_class
            except Exception as e:
                raise Exception(f"Error while adding a request handler. Please make sure it is a {InoftRequestHandler} class object : {e}")
        else:
            raise Exception(f"The following request handler is not a valid handler or do not have "
                            f"{InoftRequestHandler.__name__} as its MetaClass : {request_handler_instance_or_class}")

    def add_state_handler(self, state_handler_instance_or_class) -> None:
        if state_handler_instance_or_class is not None:
            try:
                if isinstance(state_handler_instance_or_class, type):
                    # If the variable is a class object we create an instance of the class
                    handler_bases_parent_classes = state_handler_instance_or_class.__bases__
                    state_handler_instance_or_class = state_handler_instance_or_class()
                else:
                    # If the variable is a class instance
                    handler_bases_parent_classes = state_handler_instance_or_class.__class__.__bases__

                if InoftStateHandler in handler_bases_parent_classes:
                    state_handler_instance_or_class.handler_input = self.handler_input
                    # We set a reference to the skill handler_input in each handler so that it can use it with its HandlerInputWrapper
                    self.state_handlers_chain[state_handler_instance_or_class.__class__.__name__] = state_handler_instance_or_class
            except Exception as e:
                raise Exception(f"Error while adding a state handler. Please make sure it is a {InoftStateHandler} class object : {e}")
        else:
            raise Exception(f"The following state handler is not a valid handler or do not have "
                            f"{InoftStateHandler.__name__} as its MetaClass : {state_handler_instance_or_class}")

    def set_default_fallback_handler(self, default_fallback_handler_instance_or_class) -> None:
        if default_fallback_handler_instance_or_class is not None:
            try:
                if isinstance(default_fallback_handler_instance_or_class, type):
                    # If the variable is a class object we create an instance of the class
                    handler_bases_parent_classes = default_fallback_handler_instance_or_class.__bases__
                    default_fallback_handler_instance_or_class = default_fallback_handler_instance_or_class()
                else:
                    # If the variable is a class instance
                    handler_bases_parent_classes = default_fallback_handler_instance_or_class.__class__.__bases__

                if InoftDefaultFallback in handler_bases_parent_classes:
                    default_fallback_handler_instance_or_class.handler_input = self.handler_input
                    # We set a reference to the skill handler_input in each handler so that it can use it with its HandlerInputWrapper
                    self.default_fallback_handler = default_fallback_handler_instance_or_class
            except Exception as e:
                raise Exception(f"Error while adding a request handler. Please make sure it is a {InoftDefaultFallback} class object : {e}")
        else:
            raise Exception(f"The following fallback handler is not a valid handler or do not have "
                            f"{InoftDefaultFallback.__name__} as its MetaClass : {default_fallback_handler_instance_or_class}")

    def process_request(self):
        output_event = None
        handler_to_use = None
        handler_is_an_alone_callback_function = False
        handler_is_an_audioplayer_handlers_group = False
        handler_is_a_then_state_handler = False
        handler_is_a_request_handler = False

        # Steps of priority

        # First, if the request is an interactive option made by the user
        if self.handler_input.is_option_select_request:
            # todo: rename is_option_select_request to need_to_be_handled_by_callback
            infos_callback_function_to_use = self.handler_input.interactivity_callback_functions.get(
                self.handler_input.selected_option_identifier).to_safedict(default=None)

            if infos_callback_function_to_use is not None:
                from inoft_vocal_framework.skill_builder.utils import get_function_or_class_from_file_and_path
                handler_to_use = get_function_or_class_from_file_and_path(
                    file_filepath=infos_callback_function_to_use.get("file_filepath_containing_callback").to_str(),
                    path_qualname=infos_callback_function_to_use.get("callback_function_path").to_str())

                if handler_to_use is not None:
                    handler_is_an_alone_callback_function = True

        # Second, Alexa Audio Player
        if self.handler_input.is_alexa_v1:
            if self.handler_input.alexaHandlerInput.context.audioPlayer.token is not None:
                last_used_audioplayer_handlers_group_infos = self.handler_input.alexaHandlerInput.get_last_used_audioplayer_handlers_group()
                from inoft_vocal_framework.skill_builder.utils import get_function_or_class_from_file_and_path
                audioplayer_handlers_group_class_type = get_function_or_class_from_file_and_path(
                    file_filepath=last_used_audioplayer_handlers_group_infos.get("fileFilepathContainingClass").to_str(),
                    path_qualname=last_used_audioplayer_handlers_group_infos.get("classPath").to_str())

                if audioplayer_handlers_group_class_type is not None:
                    raise_if_value_not_in_list(value=InoftHandlersGroup, list_object=list(audioplayer_handlers_group_class_type.__bases__),
                                               variable_name="audioplayer_handlers_group_class_type")

                    class_kwargs = last_used_audioplayer_handlers_group_infos.get("classKwargs").to_dict()
                    class_kwargs["parent_handler"] = self
                    handler_to_use = audioplayer_handlers_group_class_type(**class_kwargs)

                    # When using an audioplayer handlers group, we will call its handle function (it will try every one of its
                    # handler, until he found one that return an output). If the output is None (no function out of every function
                    # of the audioplayer handlers group has returned something), then we will set back the handler_to_use to None,
                    # so that the others more traditional handlers can have a chance to be use (if we did not do that, and that
                    # no event was returned, the response would be the default fallback right away).
                    # The reason we do all of that, is that if the AudioPlayer object is present, but not in a state that is
                    # supported by the app trough a can_handle function (like if it is stopped, and no can_handle function of
                    # any class is triggered by the current state of the AudioPlayer), then we will not be able to give an
                    # interactive experience with the AudioPlayer, which translate into our output_event being None.
                    output_event = handler_to_use.handle()
                    if output_event is not None:
                        handler_is_an_audioplayer_handlers_group = True
                    else:
                        handler_to_use = None

        # Third, if the invocation is a new session, and a session can be resumed, we resume the last intent of the previous session
        if self.handler_input.is_invocation_new_session is True and self.handler_input.session_been_resumed is True:
            last_intent_handler_class_key_name = self.handler_input.session_remember("lastIntentHandler")
            if last_intent_handler_class_key_name in self.request_handlers_chain.keys():
                handler_to_use = self.request_handlers_chain[last_intent_handler_class_key_name]
            elif last_intent_handler_class_key_name in self.state_handlers_chain.keys():
                handler_to_use = self.state_handlers_chain[last_intent_handler_class_key_name]

        # Fourth, loading of the then_state in the session
        if handler_to_use is None:
            last_then_state_class_name = self.handler_input.remember_session_then_state()
            if last_then_state_class_name is not None:
                if last_then_state_class_name in self.state_handlers_chain.keys():
                    handler_to_use = self.state_handlers_chain[last_then_state_class_name]
                    handler_is_a_then_state_handler = True
                    self.handler_input.forget_session_then_state()
                else:
                    logging.warning(f"A thenState class name ({last_then_state_class_name}) was not None and has"
                                    f" not been found in the available classes : {self.state_handlers_chain}")

        # Fifth, classical requests handlers
        if handler_to_use is None:
            for request_handler in self.request_handlers_chain.values():
                if request_handler.can_handle() is True:
                    handler_to_use = request_handler
                    handler_is_a_request_handler = True
                    break
                else:
                    logging.debug(f"Not handled by : {request_handler.__class__}")

        if handler_to_use is not None:
            if handler_is_an_alone_callback_function is True:
                output_event = handler_to_use(self.handler_input, self.handler_input.selected_option_identifier)
                if output_event is not None:
                    logging.debug(f"Successfully resumed by {handler_to_use} which returned {output_event}")
                else:
                    logging.info(f"A callback function has been found {handler_to_use}. But nothing was returned,"
                                 f" did you called the return self.to_platform_dict() function ?")

            if output_event is None and self.handler_input.session_been_resumed is True:
                logging.debug(f"Handled and resumed by : {handler_to_use.__class__}")
                output_event = handler_to_use.handle_resume()

            if output_event is None:
                logging.debug(f"Handled classically by : {handler_to_use.__class__}")
                output_event = handler_to_use.handle()
                if handler_is_a_then_state_handler is True and output_event is None:
                    # If the handle function of the a then_state handler do not return anything, we call its fallback
                    # function, and we set back the then_state to the session attributes (we do so before calling
                    # the fallback, in case the fallback function changed the then_state)
                    self.handler_input.memorize_session_then_state(state_handler_class_type_or_name=last_then_state_class_name)
                    output_event = handler_to_use.fallback()

        if output_event is not None:
            if handler_is_an_alone_callback_function is False and handler_is_an_audioplayer_handlers_group is False:
                # If the response is handled by a function, we do not save it as the last intent (we cannot actually,
                # and it would not make sense anyway). So we will keep the previous last intent as the new last intent.
                # We do the same thing it is handler by an audioplayer handlers group (since it is used to handle interactions too)
                self.handler_input.memorize_session_last_intent_handler(handler_class_type_instance_name=handler_to_use)
        else:
            logging.debug(f"Handler by default fallback : {self.default_fallback_handler}")
            output_event = self.default_fallback_handler.handle()

        self.handler_input.save_attributes_if_need_to()

        print(f"output_event = {output_event}")
        wrapped_output_event = LambdaResponseWrapper(response_dict=output_event).get_wrapped(handler_input=self.handler_input)
        return wrapped_output_event

    def check_everything_implemented(self):
        if self.default_fallback_handler is None:
            raise Exception(f"A skill must have a {InoftDefaultFallback} handler set with the {self.set_default_fallback_handler} function.")

    def handle_any_platform(self, event: dict, context: dict):
        print(f"Event = {json_dumps(event)}\nContext = {context}")
        self.check_everything_implemented()
        event_safedict = SafeDict(classic_dict=event)

        # The 'rawPath' is for ApiGatewayV2, use the key 'resource' (without the comma) if using ApiGatewayV1
        if event_safedict.get("rawPath").to_str() == "/googleAssistantDialogflowV1":
            # A google-assistant or dialogflow request always pass trough an API gateway
            self.handler_input.is_dialogflow_v1 = True
            event = NestedObjectToDict.get_dict_from_json(event_safedict.get("body").to_str())

        elif event_safedict.get("rawPath").to_str() == "/samsungBixbyV1":
            # A samsung bixby request always pass trough an API gateway
            self.handler_input.is_bixby_v1 = True
            event = NestedObjectToDict.get_dict_from_json(event_safedict.get("body").to_str())

        elif "amzn1." in event_safedict.get("context").get("System").get("application").get("applicationId").to_str():
            # Alexa always go last, since it do not pass trough an api resource, its a less robust identification than the other platforms.
            self.handler_input.is_alexa_v1 = True
        else:
            from inoft_vocal_framework.messages import ERROR_PLATFORM_NOT_SUPPORTED
            raise Exception(ERROR_PLATFORM_NOT_SUPPORTED)

        self.handler_input.load_event(event=event)
        return self.process_request()

    @property
    def request_handlers_chain(self) -> dict:
        return self._request_handlers_chain

    @property
    def state_handlers_chain(self) -> dict:
        return self._state_handlers_chain

    @property
    def default_fallback_handler(self):
        return self._default_fallback_handler

    @default_fallback_handler.setter
    def default_fallback_handler(self, default_fallback_handler) -> None:
        self._default_fallback_handler = default_fallback_handler

    @property
    def handler_input(self) -> HandlerInput:
        return self._handler_input

    @property
    def default_session_data_timeout(self):
        return self._handler_input.default_session_data_timeout

    @default_session_data_timeout.setter
    def default_session_data_timeout(self, default_session_data_timeout: int) -> None:
        if not isinstance(default_session_data_timeout, int):
            raise Exception(f"default_session_data_timeout was type {type(default_session_data_timeout)} which is not valid value for his parameter.")
        self.handler_input._default_session_data_timeout = default_session_data_timeout



from inoft_vocal_framework.speechs.ssml_builder_core import SpeechsList, Speech

INTERACTION_TYPE_WELCOME = "welcome"
MSGS_WELCOME = SpeechsList(id="1", speechs=[
    Speech().add_text("Bienvenue dans le jeu ! Votre but seras simple, atteindre la fin ! Prêt à jouer ?").set_prob(1),
])

MSGS_YES = SpeechsList("2", [Speech().add_text("Très bien ! ").set_prob(1)])



import random


def safe_get_element_in_dict(dict_object: dict, key_to_get: str):
    if key_to_get in dict_object.keys():
        return dict_object[key_to_get]
    else:
        return None

def pick_one_randomly_from_list(list_to_pick_from: list):
    return list_to_pick_from[random.randint(0, len(list_to_pick_from) - 1)]

def pick_msg(speechs_list: list) -> str:
    from inoft_vocal_framework.speechs.ssml_builder_core import SpeechsList
    if isinstance(speechs_list, list):
        if len(speechs_list) > 0:
            # I'm treating the probabilities so that in the data files,
            # i can simply use pseudo probabilities that do not add up to 1
            speech_texts = list()
            sum_probabilities = 0
            for speech_item in speechs_list:
                sum_probabilities += float(speech_item.probability_value)
                # Important to use float and not int
                speech_texts.append(speech_item.speech)

            probabilities_added_to_one = 0
            probabilities = list()
            for speech_item in speechs_list:
                probability_on_one = speech_item.probability_value / sum_probabilities
                probabilities.append(probability_on_one)
                probabilities_added_to_one += probability_on_one

            # We need for the probabilities to exactly sum up to one in order to work with numpy
            if probabilities_added_to_one > 0 and probabilities_added_to_one != 1 and len(probabilities) > 0:
                probability_to_remove_or_add_to_each_element = (probabilities_added_to_one - 1) / len(probabilities)
                for probability_value_of_element in probabilities:
                    probability_value_of_element += probability_to_remove_or_add_to_each_element

            picked_elements_list = random.choices(population=speech_texts, weights=probabilities)
            if picked_elements_list is not None and len(picked_elements_list) > 0:
                return picked_elements_list[0]
            else:
                return None
        else:
            raise Exception(f"No speech item are in the current speech list : {speechs_list}")
    else:
        raise Exception(f"The messages were not a list object but {speechs_list}")

def pick_msg_old(msgs) -> str:
    """ The key of a dict element is the message, and the value its probabilities """
    from inoft_vocal_framework.speechs.ssml_builder_core import SpeechsList

    if isinstance(msgs, SpeechsList):
        msgs_dict = msgs.speechs_list
    elif isinstance(msgs, dict):
        msgs_dict = msgs
    else:
        raise Exception(f"The messages object was neither a SpeechsList or a dict but was {msgs}")

    if msgs_dict is not None:
        # I'm treating the probabilities so that in the data files,
        # i can simply use pseudo probabilities that do not add up to 1
        sum_probabilities = 0
        for probability_of_element in msgs_dict.values():
            sum_probabilities += float(probability_of_element)
            # Important to use float and not int

        probabilities_added_to_one = 0
        probabilities = list()
        for probability_of_element in msgs_dict.values():
            probability_on_one = probability_of_element / sum_probabilities
            probabilities.append(probability_on_one)
            probabilities_added_to_one += probability_on_one

        # We need for the probabilities to exactly sum up to one in order to work with numpy
        if probabilities_added_to_one > 0 and probabilities_added_to_one != 1 and len(probabilities) > 0:
            probability_to_remove_or_add_to_each_element = (probabilities_added_to_one - 1) / len(probabilities)
            for probability_value_of_element in probabilities:
                probability_value_of_element += probability_to_remove_or_add_to_each_element

        print(probabilities)
        picked_elements_list = random.choices(population=list(msgs_dict.keys()), weights=probabilities)
        if picked_elements_list is not None and len(picked_elements_list) > 0:
            return picked_elements_list[0]
        else:
            return None
    else:
        return "The variable send to the pick_msg was not in the form of a dict."



from inoft_vocal_framework.platforms_handlers.nested_object_to_dict import NestedObjectToDict
from inoft_vocal_framework.safe_dict import SafeDict


class IntentSlot:
    def __init__(self):
        pass

class Intent:
    json_key = "intent"

    def __init__(self):
        self._name = str()
        self._confirmationStatus = str()
        self._slots = SafeDict()

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        if not isinstance(name, str):
            raise Exception(f"name was type {type(name)} which is not valid value for his parameter.")
        self._name = name

    @property
    def confirmationStatus(self) -> str:
        return self._confirmationStatus

    @confirmationStatus.setter
    def confirmationStatus(self, confirmationStatus: str) -> None:
        if not isinstance(confirmationStatus, str):
            raise Exception(f"confirmationStatus was type {type(confirmationStatus)} which is not valid value for his parameter.")
        self._confirmationStatus = confirmationStatus

    @property
    def slots(self) -> SafeDict:
        if isinstance(self._slots, dict):
            self._slots = SafeDict(self._slots)
        return self._slots

    @slots.setter
    def slots(self, slots: dict) -> None:
        if not isinstance(slots, dict):
            raise Exception(f"slots was type {type(slots)} which is not valid value for his parameter.")
        self._slots = SafeDict(slots)

class Request:
    json_key = "request"

    LaunchRequestKeyName = "LaunchRequest"
    IntentRequestKeyName = "IntentRequest"
    SessionEndedRequestKeyName = "SessionEndedRequest"

    def __init__(self):
        # General for LaunchRequest, IntentRequest and SessionEndedRequest
        self._type = str()
        self._requestId = str()
        self._timestamp = str()
        self._locale = str()

        # Only for an IntentRequest
        self._dialogState = None
        self._intent = Intent()

    def process_and_set_json_to_object(self, stringed_request_json_dict: str):
        NestedObjectToDict.process_and_set_json_to_object(object_class_to_set_to=self,
                                                          request_json_dict_stringed_dict_or_list=stringed_request_json_dict)

    def is_launch_request(self) -> bool:
        if self.type == self.LaunchRequestKeyName:
            return True
        else:
            return False

    def is_in_intent_names(self, intent_names_list) -> bool:
        if self.type == self.IntentRequestKeyName:
            if isinstance(intent_names_list, list):
                if self.intent.name in intent_names_list:
                    return True
            elif isinstance(intent_names_list, str):
                if self.intent.name == intent_names_list:
                    return True
            else:
                raise Exception(f"The intent_names_list must be a list or a str, but it was"
                                f" a {type(intent_names_list)} object : {intent_names_list}")
        return False

    def is_in_request_types(self, request_types_list) -> bool:
        if isinstance(request_types_list, list):
            if self.type in request_types_list:
                return True
        elif isinstance(request_types_list, str):
            if self.type == request_types_list:
                return True
        else:
            raise Exception(f"The request_types_list must be a list or a str, but it was"
                            f" a {type(request_types_list)} object : {request_types_list}")
        return False

    def get_intent_slot_value(self, slot_key: str, default=None):
        slot_value = self.intent.slots.get(dict_key=slot_key).get("value").to_any(default=default)
        # In Alexa, a ? as an arg value, means that it has not been specified
        return default if slot_value == "?" else slot_value

    def do_not_include(self) -> bool:
        if self.type is not None and self.type not in [self.LaunchRequestKeyName, self.IntentRequestKeyName, self.SessionEndedRequestKeyName]:
            raise Exception(f"The request type '{self.type}' is not None or any of the supported types.")

        return False
        if (self._type == str()
        or self._requestId == str()
        or self._timestamp == str()
        or self._locale == str()):
            return True
        else:
            return False

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, type_value: str) -> None:
        if not isinstance(type_value, str):
            raise Exception(f"type_value must be a str object : {type_value}")
        self._type = type

    @property
    def requestId(self):
        return self._requestId

    @requestId.setter
    def requestId(self, request_id: str) -> None:
        if not isinstance(request_id, str):
            raise Exception(f"request_id must be a str object : {request_id}")
        self._requestId = request_id

    @property
    def timestamp(self):
        return self._timestamp

    @timestamp.setter
    def timestamp(self, timestamp: str) -> None:
        if not isinstance(timestamp, str):
            raise Exception(f"timestamp must be a str object : {timestamp}")
        self._timestamp = timestamp

    @property
    def locale(self):
        return self._locale

    @locale.setter
    def locale(self, locale: str) -> None:
        if not isinstance(locale, str):
            raise Exception(f"locale must be a str object : {locale}")
        self._locale = locale

    @property
    def dialogState(self):
        return self._dialogState

    @dialogState.setter
    def dialogState(self, dialog_state: str) -> None:
        if not isinstance(dialog_state, str):
            raise Exception(f"dialog_state must be a str object : {dialog_state}")
        self._dialogState = dialog_state

    @property
    def intent(self):
        return self._intent

    @intent.setter
    def intent(self, intent: str) -> None:
        if not isinstance(intent, str):
            raise Exception(f"intent must be a str object : {intent}")
        self._intent = intent

    def to_dict(self) -> dict:
        dict_object = NestedObjectToDict.get_dict_from_nested_object(object_to_process=self,
                                                                     key_names_identifier_objects_to_go_into=["json_key"])
        dict_object["version"] = "1.0"
        dict_object["sessionAttributes"] = dict()
        return dict_object



import inspect
import os
import zipfile
from pathlib import Path

import boto3
import botocore
import click
from botocore.exceptions import ClientError
from click import ClickException

import inoft_vocal_framework
from inoft_vocal_framework.cli.cli_cache import CliCache
from inoft_vocal_framework.cli.core.core import Core
from inoft_vocal_framework.skill_builder.skill_settings import Settings


class DeployHandler(Core):
    def __init__(self):
        super().__init__()

    def handle(self):
        changed_root_folderpath = False
        app_project_root_folderpath = CliCache.cache().get("last_app_project_root_folderpath").to_str(default=None)

        def prompt_user_to_select_folderpath():
            return click.prompt(text="What is the root folder path of your project ? "
                                     "This is the default if you do not write anything :",
                                default=Path(os.path.dirname(os.path.realpath(inoft_vocal_framework.__file__))).parent)


        if app_project_root_folderpath is None:
            app_project_root_folderpath = prompt_user_to_select_folderpath()
            changed_root_folderpath = True
        while not os.path.exists(app_project_root_folderpath):
            if click.confirm(text="The root folder path of the project has not been found."
                                  "Do you want to select a new folderpath ? Otherwise the CLI will close."):
                app_project_root_folderpath = prompt_user_to_select_folderpath()
                changed_root_folderpath = True
            else:
                exit(200)

        if changed_root_folderpath is True:
            CliCache.cache().put("last_app_project_root_folderpath", app_project_root_folderpath)
            CliCache.save_cache_to_yaml()
            print(f"Saved the folderpath of your project for {click.style(text='faster load next time ;)', fg='blue')}")

        self.settings.find_load_settings_file(root_folderpath=app_project_root_folderpath)

        # I cannot create a variable that contain the deployment settings, otherwise it will be hell with the resets of the safedict.
        bucket_name = self.settings.settings.get_set("deployment", {}).get("s3_bucket_name").to_str(default=None)
        if bucket_name is None:
            bucket_name = click.prompt("What will be your S3 bucket name to host your deployment files ?", type=str)
            self.settings.settings.get_set("deployment", {}).put("s3_bucket_name", bucket_name).reset_navigated_dict()

        lambda_name = self.settings.settings.get_set("deployment", {}).get("lambda_name").to_str(default=None)
        if lambda_name is None:
            lambda_name = click.prompt("What will be your Lambda (which act as the 'server' for your app) name ?", type=str)
            self.settings.settings.get_set("deployment", {}).put("lambda_name", lambda_name).reset_navigated_dict()

        self.settings.save_settings()
        print(f"Saved your settings to your app_settings file at :"
              f"{click.style(text=str(self.settings.last_settings_filepath), fg='yellow')}")

        # todo: ask and check lambda handler file/function path

        DeployHandler().deploy(
            app_project_root_folderpath=app_project_root_folderpath,
            bucket_name=bucket_name,
            lambda_name=lambda_name,
            lambda_handler="app.lambda_handler",
            upload_zip=True,
        )

    def deploy(self, app_project_root_folderpath: str, bucket_name: str, lambda_name: str, lambda_handler: str,
               upload_zip: bool = True, app_project_existing_zip_filepath: str = None,
               lambda_description: str = "Inoft Vocal Framework Deployment",
               lambda_timeout_seconds=30, lambda_memory_size=512, publish=True, runtime="python3.7"):

        if self.settings.settings_loaded is False:
            self.settings.find_load_settings_file(root_folderpath=app_project_root_folderpath)

        # Make sure this isn't already deployed.
        """deployed_versions = self.get_lambda_function_versions(lambda_name)
        if len(deployed_versions) > 0:
            print(f"This lambda function for this application is {click.style('already deployed', fg='green')}" +
                  f" If some part of the full deployment failed or you changed some resource name, and you want to do it again, type Y."
                  f" Otherwise, use the {click.style('inoft update', bold=True)} command")
            if not click.confirm("Do you want to redo the deployment ?"):
                print(f"Ok ! Remember to use the {click.style('inoft update', bold=True)} command !")
                exit()
        """

        zip_filepath, upload_success = None, None
        if upload_zip is True:
            if app_project_existing_zip_filepath is None:
                # Create the Lambda Zip
                zip_filepath = self.create_package(app_folder_path=app_project_root_folderpath)
                # todo: do not include any venv or sam folder when deploying
            else:
                if not os.path.isfile(app_project_existing_zip_filepath):
                    raise Exception(f"Existing app project zip file do not exist at filepath : {app_project_existing_zip_filepath}")
                zip_filepath = app_project_existing_zip_filepath

            # Upload it to S3
            click.echo("Uploading the app zip file to S3")
            upload_success = self.upload_to_s3(filepath=zip_filepath, object_key_name=Path(zip_filepath).name,
                                               bucket_name=bucket_name, region_name="eu-west-3")
            if not upload_success:
                raise ClickException("Unable to upload to S3. Look in the logs what caused the errors,"
                                     " and modify your app_settings file in order to fix the issue."
                                     "\nCommon issues include :"
                                     "\n - Trying to use a S3 bucket name that is not available."
                                     "\n - Trying to use invalid characters in the bucket name (like - )"
                                     "\n - Having modified the app_settings file, but not have saved it with CTRL+S")
            click.echo("Uploading completed.")

        try:
            lambda_arn = self.get_lambda_function_arn(function_name=lambda_name)
            print(f"Using the existing lambda function {lambda_name}")
            if upload_success is True:
                self.update_lambda_function_code(lambda_arn=lambda_arn, object_key_name=Path(zip_filepath).name, bucket_name=bucket_name)
                click.echo(f"Updated the code of the lambda with arn {lambda_arn}")

        except botocore.exceptions.ClientError:
            try:
                lambda_arn = self.create_lambda_function(bucket=bucket_name, s3_key=Path(zip_filepath).name, function_name=lambda_name,
                                                         handler=lambda_handler, description=lambda_description,
                                                         timeout=lambda_timeout_seconds, memory_size=lambda_memory_size,
                                                         runtime=runtime, publish=publish)
                print(f"Created the lambda function : {lambda_name}")
            except Exception as e:
                print(f"Error while updating the function by giving it a S3 path. Trying to update the"
                      f"function with a byte_stream of the zip file of the deployment package : {e}")

                with open(zip_filepath, mode="rb") as file_stream:
                    byte_stream = file_stream.read()

                lambda_arn = self.create_lambda_function(local_zip=byte_stream, function_name=lambda_name,
                                                         handler=lambda_handler, description=lambda_description,
                                                         timeout=lambda_timeout_seconds, memory_size=lambda_memory_size,
                                                         runtime=runtime, publish=publish)
                print(f"Created the lambda function : {lambda_name}")

        # Create and configure the API Gateway
        need_to_create_api = False
        api_gateway_id = self.settings.settings.get("deployment").get("apiGatewayId").to_str(default=None)
        if api_gateway_id is None or self.api_gateway_v2_url(api_gateway_id) is None:
            api_id = self.create_api_gateway(lambda_arn=lambda_arn, lambda_name=lambda_name)
            print(f"Created a new {click.style(text='API Gateway', bold=True)}")
            self.settings.settings.get("deployment").put("apiGatewayId", api_id)
            self.settings.save_settings()
        else:
            print(f"Using the existing ApiGatewayV2 with id {click.style(text=api_gateway_id, bold=True, fg='green')}")

        return True

        # Remove the project zip from S3.
        if not source_zip:
            self.remove_uploaded_zip()


    @staticmethod
    def create_package(app_folder_path: str) -> str:
        # todo: fix issue where we have the following printed for each single file (most likely come from the fact that i do not delete the old zip before creating a new one)
        #  return self._open_to_write(zinfo, force_zip64=force_zip64)
        #  c:\users\labourdette\anaconda3\lib\zipfile.py:1473: UserWarning: Duplicate name: 'inoft_vocal_framework/utils/__init__.py'

        archive_destination_filepath = os.path.join(Path(app_folder_path).parent, f"{Path(app_folder_path).name}.zip")
        click.echo(f"Making an archive from all the files and folders in {app_folder_path} to {archive_destination_filepath}")

        folders_names_to_excludes = [".aws-sam", ".idea", "__pycache__", "venv"]

        # Create a ZipFile Object
        with zipfile.ZipFile(archive_destination_filepath, "w") as zip_object:
            # Include the projects files
            for root_dirpath, dirs, filenames in os.walk(app_folder_path, topdown=True):
                # The topdown arg allow use to modify the dirs list in the walk, and so we can easily exclude folders.
                dirs[:] = [dirpath for dirpath in dirs if Path(dirpath).name not in folders_names_to_excludes]

                relative_root_dirpath = root_dirpath.replace(app_folder_path, "")
                for filename in filenames:
                    zip_object.write(filename=os.path.join(root_dirpath, filename),
                                     arcname=os.path.join(relative_root_dirpath, filename))

            # todo: when doing a redeploy, check that the lambda layer used, is the right layer for the current framework version

            # Then the inoft_vocal_framework himself (he is not included in the lambda layers)
            # I do not include the framework in the lambda layer, because it would be weird for someone to update the framework,
            # do a redeploy, and not have upgraded its deploy. Where as we can check if its layer is the right layer for the version
            # of his framework. And mostly because it would be annoying for me to recreate a new lambda layer on each update of the
            # framework, and that he is light enough to be included in the package and the increase in upload time to not be noticeable ;)
            inoft_vocal_framework_folder_path = os.path.dirname(os.path.realpath(inoft_vocal_framework.__file__))
            for root_dirpath, dirs, filenames in os.walk(inoft_vocal_framework_folder_path):
                relative_dirpath_with_root_included = root_dirpath.replace(str(Path(inoft_vocal_framework_folder_path).parent), "")
                # We only replace the parent of the framework folder path, because we want it to be in its inoft_vocal_framework folder.
                for filename in filenames:
                    zip_object.write(filename=os.path.join(root_dirpath, filename),
                                     arcname=os.path.join(relative_dirpath_with_root_included, filename))


        # from shutil import make_archive
        # zip_filepath = make_archive(base_name=archive_destination_filepath_without_extension, format="zip", root_dir=app_folder_path)
        print(f"Archive completed at {click.style(text=archive_destination_filepath, bold=True)}")

        # Warn if this is too large for Lambda.
        file_stats = os.stat(archive_destination_filepath)
        if file_stats.st_size > 52428800:
            click.echo("Warning: Application zip package is likely to be too large for AWS Lambda. Try to make it smaller")

        return archive_destination_filepath


if __name__ == "__main__":
    DeployHandler().handle()



import time

from boto3.session import ResourceNotExistsError
from inoft_vocal_framework.exceptions import raise_if_variable_not_expected_type

from inoft_vocal_framework.databases.dynamodb.dynamodb_core import DynamoDbCoreAdapter
from inoft_vocal_framework.safe_dict import SafeDict
from inoft_vocal_framework.speechs.ssml_builder_core import Speech, SpeechsList


class DynamoDbAttributesAdapter(DynamoDbCoreAdapter):
    def __init__(self, table_name: str, region_name: str, primary_key_name="id", create_table=True,
                 persistent_attributes_key_name="persistentAttributes", smart_session_attributes_key_name="smartSessionAttributes"):
        super().__init__(table_name=table_name, region_name=region_name)

        self.persistent_attributes_key_name = persistent_attributes_key_name
        self.smart_session_attributes_key_name = smart_session_attributes_key_name

        self._fetchedData = None
        self._last_user_id = None

    # todo: create an option of bandwith optimization by loading the
    #  persistent attributes in the session attributes at the start of the session

    def _fetch_attributes(self, user_id: str) -> SafeDict:
        try:
            from boto3.dynamodb.types import TypeSerializer
            response = self.dynamodb.get_item(TableName=self.table_name, Key={"id": TypeSerializer().serialize(user_id)}, ConsistentRead=True)
            if "Item" in response:
                return SafeDict(self.utils.python_to_dynamodb(response["Item"]))
            else:
                return SafeDict()
        except ResourceNotExistsError:
            raise Exception(f"DynamoDb table {self.table_name} do not exist or in the process"
                            "of being created. Failed to get attributes from DynamoDb table.")
        except Exception as e:
            raise Exception(f"Failed to retrieve attributes from DynamoDb table."
                            f"Exception of type {type(e).__name__} occurred: {str(e)}")

    @property
    def fetchedData(self) -> SafeDict:
        if self._fetchedData is None:
            self._fetchedData = self._fetch_attributes(user_id=self.last_user_id)
        return self._fetchedData

    def get_field(self, user_id: str, field_key: str):
        self.last_user_id = user_id
        return self.fetchedData.get(field_key).to_any()

    def get_smart_session_attributes(self, user_id: str, session_id: str, timeout_seconds: int) -> (SafeDict, bool):
        # If the value from get_field is of dict or list type, the SafeDict will be populated, otherwise it will be empty without errors.
        timeout_expired = True

        last_session_id = self.get_field(user_id=user_id, field_key="lastSessionId")
        if session_id == last_session_id:
            timeout_expired = False
        else:
            last_interaction_time = self.get_field(user_id=user_id, field_key="lastInteractionTime")
            if last_interaction_time is not None and time.time() <= last_interaction_time + timeout_seconds:
                timeout_expired = False

        if timeout_expired is False:
            return SafeDict(self.get_field(user_id=user_id, field_key=self.smart_session_attributes_key_name)), True
        else:
            return SafeDict(), False

    # todo: make it so that the smart session data timeout can be customized for each user, and is saved and retrieved from the db at each invocation
    def get_persistent_attributes(self, user_id: str) -> SafeDict:
        # If the value from get_field is of dict or list type, the SafeDict will be populated, otherwise it will be empty without errors.
        return SafeDict(self.get_field(user_id=user_id, field_key=self.persistent_attributes_key_name))

    def save_attributes(self, user_id: str, session_id: str, smart_session_attributes: dict, persistent_attributes: dict) -> None:
        self.last_user_id = user_id

        if not isinstance(smart_session_attributes, dict):
            print(f"smart_session_attributes has been set to empty dict because was not of dict type but {smart_session_attributes}")
            smart_session_attributes = dict()
        if not isinstance(persistent_attributes, dict):
            print(f"persistent_attributes has been set to empty dict because was not of dict type but {persistent_attributes}")
            persistent_attributes = dict()

        item_dict = {
            "id": user_id,
            "lastSessionId": session_id,
            "lastInteractionTime": round(time.time()),  # No need to store milliseconds for the last interaction time
            self.smart_session_attributes_key_name: smart_session_attributes,
            self.persistent_attributes_key_name: persistent_attributes
        }

        try:
            print(f"Saving attributes : {item_dict}")
            # todo: fix issue where bool are converted to int (it happens when the data is send to dynamodb, but the weird stuff is
            #  that it happens in the framework, but not if i do the same thing and send the same data in a python console... *-*
            self.dynamodb.put_item(TableName=self.table_name, Item=self.utils.python_to_dynamodb(item_dict))
        except ResourceNotExistsError:
            raise Exception(f"DynamoDb table {self.table_name} doesn't exist. Failed to save attributes to DynamoDb table.")
        except Exception as e:
            raise Exception(f"Failed to save attributes to DynamoDb table. Exception of type {type(e).__name__} occurred: {str(e)}")

    def delete_attributes(self, user_id: str):
        """ Deletes attributes from table in Dynamodb resource. """
        self.last_user_id = user_id
        try:
            self.dynamodb.delete_item(TableName=self.table_name, Key={self.primary_key_name: user_id})
        except ResourceNotExistsError:
            raise Exception(f"DynamoDb table {self.table_name} doesn't exist. Failed to delete attributes from DynamoDb table.")
        except Exception as e:
            raise Exception(f"Failed to delete attributes in DynamoDb table. Exception of type {type(e).__name__} occurred: {str(e)}")

    @property
    def last_user_id(self) -> str:
        return self._last_user_id

    @last_user_id.setter
    def last_user_id(self, last_user_id: str) -> None:
        if not isinstance(last_user_id, str):
            raise Exception(f"last_user_id was type {type(last_user_id)} which is not valid value for his parameter.")
        self._last_user_id = last_user_id


class DynamoDbMessagesAdapter(DynamoDbCoreAdapter):
    def __init__(self, table_name: str, region_name: str, is_admin_mode=False,
                 persistent_attributes_key_name="persistentAttributes", smart_session_attributes_key_name="smartSessionAttributes"):
        super().__init__(table_name=table_name, region_name=region_name)

        self.persistent_attributes_key_name = persistent_attributes_key_name
        self.smart_session_attributes_key_name = smart_session_attributes_key_name

        self.is_admin_mode = is_admin_mode
        self._last_fetched_messages_list = None
        self._last_messages_list_id = None

    def _speech_dicts_to_speech_items(self, speech_dicts: list) -> list:
        output_list = list()
        for speech_dict in speech_dicts:
            output_list.append(Speech().from_dict(speech_safedict=SafeDict(speech_dict)))
        return output_list

    def get_fetched_messages_list(self, messages_list_id: str) -> list:
        if self._last_fetched_messages_list is None or messages_list_id != self.last_messages_list_id:
            try:
                response = self.dynamodb.get_item(TableName=self.table_name, Key={"id": messages_list_id}, ConsistentRead=True)
                if "Item" in response and "speechItems" in response["Item"]:
                    self._last_fetched_messages_list = self._speech_dicts_to_speech_items(self.utils.dynamodb_to_python(response["Item"]["speechItems"]))
                else:
                    self._last_fetched_messages_list = list()
            except ResourceNotExistsError:
                raise Exception(f"DynamoDb table {self.table_name} do not exist or in the process of "
                                f"being created. Failed to get messages list from DynamoDb table.")
            except Exception as e:
                raise Exception(f"Failed to retrieve messages list from DynamoDb table. "
                                f"Exception of type {type(e).__name__} occurred: {str(e)}")
        return self._last_fetched_messages_list

    def get_speechs_list(self, messages_list_id: str) -> list:
        return self.get_fetched_messages_list(messages_list_id=messages_list_id)

    @staticmethod
    def _speech_item_to_dict(speech_item: Speech) -> dict:
        return {"speech": speech_item.speech, "probability": speech_item.probability_value}

    def _speechs_list_object_to_dict(self, speechs_list_object: SpeechsList) -> dict:
        speech_items_list = list()
        for speech_item in speechs_list_object.speechs_list:
            speech_items_list.append(self._speech_item_to_dict(speech_item=speech_item))
        return {"id": speechs_list_object.interaction_type_names_list[0], "speechItems": speech_items_list}

    def post_new_speechs_list(self, speechs_list: SpeechsList):
        if self.is_admin_mode is not True:
            raise Exception(f"In order to post a new speechs list, the admin mode must be activated but was {self.is_admin_mode}")
        else:
            try:
                speechs_list_dict = self._speechs_list_object_to_dict(speechs_list_object=speechs_list)
                print(f"Posting speechs list : {speechs_list_dict}")
                self.dynamodb.put_item(TableName=self.table_name, Item=self.utils.python_to_dynamodb(speechs_list_dict))
            except ResourceNotExistsError:
                raise Exception(f"DynamoDb table {self.table_name} doesn't exist. Failed to save attributes to DynamoDb table.")
            except Exception as e:
                raise Exception(f"Failed to save speechs list to DynamoDb table. Exception of type {type(e).__name__} occurred: {str(e)}")

    def save_attributes(self, user_id: str, session_id: str, smart_session_attributes: dict, persistent_attributes: dict) -> None:
        self.last_user_id = user_id

        if not isinstance(smart_session_attributes, dict):
            print(f"smart_session_attributes has been set to empty dict because was not of dict type but {smart_session_attributes}")
            smart_session_attributes = dict()
        if not isinstance(persistent_attributes, dict):
            print(f"persistent_attributes has been set to empty dict because was not of dict type but {persistent_attributes}")
            persistent_attributes = dict()

        item_dict = {
            "id": user_id,
            "lastSessionId": session_id,
            "lastInteractionTime": round(time.time()),  # No need to store milliseconds for the last interaction time
            self.smart_session_attributes_key_name: smart_session_attributes,
            self.persistent_attributes_key_name: persistent_attributes
        }

        try:
            print(f"Saving attributes : {item_dict}")
            self.dynamodb.put_item(TableName=self.table_name, Item=self.utils.python_to_dynamodb(item_dict))
        except ResourceNotExistsError:
            raise Exception(f"DynamoDb table {self.table_name} doesn't exist. Failed to save attributes to DynamoDb table.")
        except Exception as e:
            raise Exception(f"Failed to save attributes to DynamoDb table. Exception of type {type(e).__name__} occurred: {str(e)}")

    def delete_attributes(self, user_id: str):
        """ Deletes attributes from table in Dynamodb resource. """
        self.last_user_id = user_id
        try:
            self.dynamodb.delete_item(TableName=self.table_name, Key={self.primary_key_name: user_id})
        except ResourceNotExistsError:
            raise Exception(f"DynamoDb table {self.table_name} doesn't exist. Failed to delete attributes from DynamoDb table.")
        except Exception as e:
            raise Exception(f"Failed to delete attributes in DynamoDb table. Exception of type {type(e).__name__} occurred: {str(e)}")

    @property
    def last_messages_list_id(self) -> str:
        return self._last_messages_list_id
    
    @last_messages_list_id.setter
    def last_messages_list_id(self, last_messages_list_id: str) -> None:
        raise_if_variable_not_expected_type(value=last_messages_list_id, expected_type=str, variable_name="last_messages_list_id")
        self._last_messages_list_id = last_messages_list_id

if __name__ == "__main__":
    messages_db = DynamoDbMessagesAdapter(is_admin_mode=True, table_name="test_messages", region_name="eu-west-3")
    INTERACTION_TYPE_QUESTION_DO_YOU_WANT_INFOS_ABOUT_THE_GAME = "question_do-you-want-infos-about-the-game"
    MSGS_DO_YOU_WANT_INFOS_ABOUT_THE_GAME = SpeechsList().types(INTERACTION_TYPE_QUESTION_DO_YOU_WANT_INFOS_ABOUT_THE_GAME).speechs({
        Speech().add_text("Tant pis, le projet de Polemika est cool, tu veut en savoir plus ?").set_prob(1): 1,
    })
    # messages_db.post_new_speechs_list(MSGS_DO_YOU_WANT_INFOS_ABOUT_THE_GAME)
    # print(messages_db.get_messages_of_category(category_id="question_do-you-want-infos-about-the-game").to_dict())
